"""
Abstraction layer above `tinydb`
"""

import os
import json
import tempfile
import tinydb
from tinydb import operations
from tinydb.middlewares import CachingMiddleware
from e4s_cl import logger, util
from e4s_cl.error import ConfigurationError
from e4s_cl.cf.storage import AbstractStorage, StorageRecord, StorageError

LOGGER = logger.get_logger(__name__)


class _JsonRecord(StorageRecord):
    eid_type = int

    def __init__(self, database, element, eid=None):
        super().__init__(database, eid or element.doc_id, element)

    def __str__(self):
        return json.dumps(self)

    def __repr__(self):
        return json.dumps(self)


class _JsonFileStorage(tinydb.JSONStorage):
    """Allow read-only as well as read-write access to the JSON file.
    
    TinyDB's default storage (:any:`tinydb.JSONStorage`) assumes write access to the JSON file.
    This isn't the case for system-level storage and possibly others.
    """

    def __init__(self, path, encoding=None):
        self.path = path

        try:
            super().__init__(path)
        except IOError:
            self.path = path
            self._handle = open(path, 'r', encoding=encoding)  #pylint: disable=consider-using-with
            self.readonly = True
            LOGGER.debug("'%s' opened read-only", path)
        else:
            self.readonly = False
            LOGGER.debug("'%s' opened read-write", path)

    def write(self, *args, **kwargs):
        if self.readonly:
            raise ConfigurationError(f"Cannot write to '{self.path}'",
                                     "Check that you have `write` access.")
        super().write(*args, **kwargs)


class LocalFileStorage(AbstractStorage):
    """A persistant, transactional record storage system.  
    
    Uses :py:class:`TinyDB` for both the database and the key/value store.
    
    Attributes:
        dbfile (str): Absolute path to database file.
    """

    Record = _JsonRecord

    def __init__(self, name, prefix):
        super().__init__(name)
        self._transaction_count = 0
        self._db_copy = None
        self._database = None
        self._prefix = prefix

    def __len__(self):
        return self.count()

    def __getitem__(self, key):
        record = self.get({'key': key})
        if record is not None:
            return record['value']
        raise KeyError

    def __setitem__(self, key, value):
        if self.contains({'key': key}):
            self.update({'value': value}, {'key': key})
        else:
            self.insert({'key': key, 'value': value})

    def __delitem__(self, key):
        if not self.contains({'key': key}):
            raise KeyError
        self.remove({'key': key})

    def __contains__(self, key):
        return self.contains({'key': key})

    def __iter__(self):
        for item in self.search():
            yield item['key']

    def iterkeys(self):
        for item in self.search():
            yield item['key']

    def itervalues(self):
        for item in self.search():
            yield item['value']

    def iteritems(self):
        for item in self.search():
            yield item['key'], item['value']

    def is_writable(self):
        """Check if the storage filesystem is writable.
        
        Attempts to create and delete a file in ``prefix``.
        See https://github.com/ParaToolsInc/taucmdr/issues/231.
        
        Returns:
            bool: True if a file could be created and deleted in ``prefix``, False otherwise.
        """
        self.connect_filesystem()
        if not os.access(self.prefix(), os.W_OK):
            return False
        try:
            with tempfile.NamedTemporaryFile(dir=self.prefix(),
                                             mode='w',
                                             delete=True) as tmp_file:
                tmp_file.write("Write test. Delete this file.")
        except (OSError, IOError):
            return False
        return True

    def connect_filesystem(self, *args, **kwargs):
        """Prepares the store filesystem for reading and writing."""
        if not os.path.isdir(self._prefix):
            try:
                util.mkdirp(self._prefix)
            except Exception as err:
                raise StorageError(
                    f"Failed to access {self.name} filesystem prefix '{self._prefix}': {err}"
                ) from err
            LOGGER.debug("Initialized %s filesystem prefix '%s'", self.name,
                         self._prefix)

    def disconnect_filesystem(self, *args, **kwargs):
        """Disconnects the store filesystem."""
        self.disconnect_database()

    def connect_database(self, *args, **kwargs):
        """Open the database for reading and writing."""
        if self._database is None:
            util.mkdirp(self.prefix())
            dbfile = os.path.join(self.prefix(), self.name + '.json')
            try:
                storage = CachingMiddleware(_JsonFileStorage)
                storage.WRITE_CACHE_SIZE = 0
                self._database = tinydb.TinyDB(dbfile, storage=storage)
            except IOError as err:
                raise StorageError(
                    f"Failed to access {self.name} database '{dbfile}': {err}"
                    "Check that you have `write` access") from err
            if not util.path_accessible(dbfile):
                raise StorageError(
                    f"Database file '{dbfile}' exists but cannot be read.",
                    "Check that you have `read` access")
            LOGGER.debug("Initialized %s database '%s'", self.name, dbfile)

    def disconnect_database(self, *args, **kwargs):
        """Close the database for reading and writing."""
        if self._database:
            self._database.close()
            self._database = None

    def prefix(self):
        return self._prefix

    def __str__(self):
        """Human-readable identifier for this database."""
        # pylint: disable=protected-access
        return self._database._storage.path

    def __enter__(self):
        """Initiates the database transaction."""
        # pylint: disable=protected-access
        if self._transaction_count == 0:
            self.connect_database()
            self._db_copy = self._database._storage.read()
        self._transaction_count += 1
        return self

    def __exit__(self, ex_type, value, traceback):
        """Finalizes the database transaction."""
        # pylint: disable=protected-access
        self._transaction_count -= 1
        if ex_type and self._transaction_count == 0:
            self._database._storage.write(self._db_copy)
            self._db_copy = None
            return False
        return True

    def table(self, table_name):
        self.connect_database()
        if table_name is None:
            return self._database
        return self._database.table(table_name)

    @staticmethod
    def _query(keys, match_any):
        """Construct a TinyDB query object."""

        def _and(lhs, rhs):
            return lhs & rhs

        def _or(lhs, rhs):
            return lhs | rhs

        join = _or if match_any else _and
        itr = list(keys.items())
        key, val = itr[0]
        query = (tinydb.where(key) == val)
        for key, value in itr[1:]:
            query = join(query, (tinydb.where(key) == value))
        return query

    def count(self, table_name=None):
        """Count the records in the database.
        
        Args:
            table_name (str): Name of the table to operate on.  See :any:`AbstractDatabase.table`.
            
        Returns:
            int: Number of records in the table.
        """
        return len(self.table(table_name))

    def get(self, keys, table_name=None, match_any=False):
        """Find a single record.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: return the record with that element identifier.
            * dict: return the record with attributes matching `keys`.
            * list or tuple: return a list of records matching the elements of `keys`
            * None: return None.
        
        Args:
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractDatabase.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.

        Returns:
            Record: The matching data record if `keys` was a self.Record.eid_type or dict.
            list: All matching data records if `keys` was a list or tuple.
            None: No record found or ``bool(keys) == False``.
            
        Raises:
            ValueError: Invalid value for `keys`.
        """
        table = self.table(table_name)

        if keys is None:
            return None

        if isinstance(keys, self.Record.eid_type):
            #LOGGER.debug("%s: get(eid=%r)", table_name, keys)
            element = table.get(doc_id=keys)
        elif isinstance(keys, dict) and keys:
            #LOGGER.debug("%s: get(keys=%r)", table_name, keys)
            element = table.get(self._query(keys, match_any))
        elif isinstance(keys, (list, tuple)):
            #LOGGER.debug("%s: get(keys=%r)", table_name, keys)
            return [
                self.get(key, table_name=table_name, match_any=match_any)
                for key in keys
            ]
        else:
            raise ValueError(keys)
        if element:
            return self.Record(self, element=element)
        return None

    def search(self, keys=None, table_name=None, match_any=False):
        """Find multiple records.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: return the record with that element identifier.
            * dict: return all records with attributes matching `keys`.
            * list or tuple: return a list of records matching the elements of `keys`
            * None: return all records.
        
        Args:
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractDatabase.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.

        Returns:
            list: Matching data records.
            
        Raises:
            ValueError: Invalid value for `keys`.
        """
        #LOGGER.debug("Search '%s' for '%s'", table_name, keys)
        table = self.table(table_name)
        if keys is None:
            #LOGGER.debug("%s: all()", table_name)
            return [
                self.Record(self, element=element) for element in table.all()
            ]

        if isinstance(keys, self.Record.eid_type):
            #LOGGER.debug("%s: search(eid=%r)", table_name, keys)
            element = table.get(doc_id=keys)
            return [self.Record(self, element=element)] if element else []

        if isinstance(keys, dict) and keys:
            #LOGGER.debug("%s: search(keys=%r)", table_name, keys)
            return [
                self.Record(self, element=element)
                for element in table.search(self._query(keys, match_any))
            ]

        if isinstance(keys, (list, tuple)):
            #LOGGER.debug("%s: search(keys=%r)", table_name, keys)
            result = []
            for key in keys:
                result.extend(
                    self.search(keys=key,
                                table_name=table_name,
                                match_any=match_any))
            return result

        raise ValueError(keys)

    def match(self, field, table_name=None, regex=None, test=None):
        """Find records where `field` matches `regex` or `test`.
        
        Either `regex` or `test` may be specified, not both.  
        If `regex` is given, then all records with `field` matching the regular expression are returned.
        If test is given then all records with `field` set to a value that caues `test` to return True are returned. 
        If neither is given, return all records where `field` is set to any value. 
        
        Args:
            table_name (str): Name of the table to operate on.  See :any:`AbstractDatabase.table`.
            field (string): Name of the data field to match.
            regex (string): Regular expression string.
            test: Callable returning a boolean value.  

        Returns:
            list: Matching data records.
            
        Raises:
            ValueError: Invalid value for `keys`.
        """
        table = self.table(table_name)
        if test is not None:
            #LOGGER.debug('%s: search(where(%s).test(%r))', table_name, field, test)
            return [
                self.Record(self, element=elem)
                for elem in table.search(tinydb.where(field).test(test))
            ]

        if regex is not None:
            #LOGGER.debug('%s: search(where(%s).matches(%r))', table_name, field, regex)
            return [
                self.Record(self, element=elem)
                for elem in table.search(tinydb.where(field).matches(regex))
            ]

        #LOGGER.debug("%s: search(where(%s).matches('.*'))", table_name, field)
        return [
            self.Record(self, element=elem)
            for elem in table.search(tinydb.where(field).matches(".*"))
        ]

    def contains(self, keys, table_name=None, match_any=False):
        """Check if the specified table contains at least one matching record.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: check for the record with that element identifier.
            * dict: check for the record with attributes matching `keys`.
            * list or tuple: return the equivilent of ``map(contains, keys)``.
            * None: return False.
        
        Args:
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractDatabase.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.

        Returns:
            bool: True if the table contains at least one matching record, False otherwise.
            
        Raises:
            ValueError: Invalid value for `keys`.
        """
        table = self.table(table_name)
        if keys is None:
            return False

        if isinstance(keys, self.Record.eid_type):
            #LOGGER.debug("%s: contains(eid=%r)", table_name, keys)
            return table.contains(eid=keys)

        if isinstance(keys, dict) and keys:
            #LOGGER.debug("%s: contains(keys=%r)", table_name, keys)
            return table.contains(self._query(keys, match_any))

        if isinstance(keys, (list, tuple)):
            return [
                self.contains(keys=key,
                              table_name=table_name,
                              match_any=match_any) for key in keys
            ]

        raise ValueError(keys)

    def insert(self, data, table_name=None):
        """Create a new record.
        
        If the table doesn't exist it will be created.
        
        Args:
            data (dict): Data to insert in table.
            table_name (str): Name of the table to operate on.  See :any:`AbstractDatabase.table`.
            
        Returns:
            Record: The new record.
        """
        eid = self.table(table_name).insert(data)
        record = self.Record(self, eid=eid, element=data)
        return record

    def update(self, fields, keys, table_name=None, match_any=False):
        """Update records.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: update the record with that element identifier.
            * dict: update all records with attributes matching `keys`.
            * list or tuple: apply update to all records matching the elements of `keys`.
        
        Args:
            fields (dict): Data to record.
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractDatabase.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.
            
        Raises:
            ValueError: ``bool(keys) == False`` or invaild value for `keys`.
        """
        table = self.table(table_name)
        if isinstance(keys, self.Record.eid_type):
            #LOGGER.debug("%s: update(%r, eid=%r)", table_name, fields, keys)
            table.update(fields, doc_ids=[keys])
        elif isinstance(keys, dict):
            #LOGGER.debug("%s: update(%r, keys=%r)", table_name, fields, keys)
            table.update(fields, self._query(keys, match_any))
        elif isinstance(keys, (list, tuple)):
            table.update(fields, doc_ids=keys)
        else:
            raise ValueError(keys)

    def unset(self, fields, keys, table_name=None, match_any=False):
        """Update records by unsetting fields.
        
        Update only allows you to update a record by adding new fields or overwriting existing fields. 
        Use this method to remove a field from the record.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: update the record with that element identifier.
            * dict: update all records with attributes matching `keys`.
            * list or tuple: apply update to all records matching the elements of `keys`.
        
        Args:
            fields (list): Names of fields to remove from matching records.
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractDatabase.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.
            
        Raises:
            ValueError: ``bool(keys) == False`` or invaild value for `keys`.
        """
        table = self.table(table_name)
        if isinstance(keys, self.Record.eid_type):
            for field in fields:
                #LOGGER.debug("%s: unset(%s, eid=%r)", table_name, field, keys)
                table.update(operations.delete(field), doc_ids=[keys])
        elif isinstance(keys, dict):
            for field in fields:
                #LOGGER.debug("%s: unset(%s, keys=%r)", table_name, field, keys)
                table.update(operations.delete(field),
                             self._query(keys, match_any))
        elif isinstance(keys, (list, tuple)):
            for field in fields:
                table.update(operations.delete(field), doc_ids=keys)
        else:
            raise ValueError(keys)

    def remove(self, keys, table_name=None, match_any=False):
        """Delete records.
        
        The behavior depends on the type of `keys`:
            * self.Record.eid_type: delete the record with that element identifier.
            * dict: delete all records with attributes matching `keys`.
            * list or tuple: delete all records matching the elements of `keys`.
        
        Args:
            keys: Fields or element identifiers to match.
            table_name (str): Name of the table to operate on.  See :any:`AbstractDatabase.table`.
            match_any (bool): Only applies if `keys` is a dictionary.  If True then any key 
                              in `keys` may match or if False then all keys in `keys` must match.
            
        Raises:
            ValueError: ``bool(keys) == False`` or invaild value for `keys`.
        """
        table = self.table(table_name)
        if isinstance(keys, self.Record.eid_type):
            #LOGGER.debug("%s: remove(eid=%r)", table_name, keys)
            table.remove(doc_ids=[keys])
        elif isinstance(keys, dict):
            #LOGGER.debug("%s: remove(keys=%r)", table_name, keys)
            table.remove(self._query(keys, match_any))
        elif isinstance(keys, (list, tuple)):
            table.remove(doc_ids=keys)
        else:
            raise ValueError(keys)

    def purge(self, table_name=None):
        """Delete all records.

        Args:
            table_name (str): Name of the table to operate on.  See :any:`AbstractDatabase.table`.
        """
        LOGGER.debug("%s: purge()", table_name)
        self.table(table_name).truncate()



"""
Containerless support
"""

import os
import tempfile
from pathlib import Path
from typing import List
from e4s_cl import CONTAINER_SCRIPT
from e4s_cl.util import which, run_subprocess
from e4s_cl.logger import get_logger
from e4s_cl.cf.containers import Container, BackendNotAvailableError
from e4s_cl.cf.libraries import libc_version

LOGGER = get_logger(__name__)

NAME = 'containerless'
MIMES = []


class Containerless(Container):
    """
    Containerless object
    """

    executable_name = 'bash'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.executable = which(self.__class__.executable_name)

        # pylint: disable=R1732
        # The following directory will hold symlinks to the libraries bound to
        # the container; the file will be deleted once the object is deleted
        self._lib_dir = tempfile.TemporaryDirectory()

    @property
    def script(self):
        return Path(self._lib_dir.name, Path(CONTAINER_SCRIPT).name)

    @property
    def import_dir(self):
        return Path(self._lib_dir.name)

    @property
    def import_library_dir(self):
        return Path(self._lib_dir.name, super().import_library_dir.name)

    def get_data(self):
        self.libc_v = libc_version()

        return set()

    def _setup_import(self) -> None:
        """
        Create symlinks to bound libraries in a temporary directory
        """
        for file in self.bound:
            # Abort if not a file
            if not file.origin.resolve().is_file():
                continue

            # Abort if not bound in the special dir
            try:
                rel = file.destination.relative_to(self.import_dir)
            except ValueError:
                LOGGER.debug("%s is not in %s", file.destination,
                             self.import_dir)
                continue

            link = Path(self._lib_dir.name, rel)
            os.makedirs(link.parent, exist_ok=True)
            os.symlink(file.origin.resolve(), link)

        ld_path = os.environ.get("LD_LIBRARY_PATH")

        if ld_path:
            ld_path = f"{self._lib_dir.name}{os.pathsep}{ld_path}"
        else:
            ld_path = self._lib_dir.name

        self.env["LD_LIBRARY_PATH"] = ld_path

    def run(self, command: List[str], overload: bool = True) -> int:
        if not which(self.executable):
            raise BackendNotAvailableError(self.executable)

        self._setup_import()

        return run_subprocess(command, env=self.env)


CLASS = Containerless



from os import environ
import sys
from pathlib import Path
from importlib import import_module
from shlex import split
import tests
from e4s_cl import config
from e4s_cl.cf import launchers
from e4s_cl.cf.launchers import (
    Parser,
    filter_arguments,
    get_reserved_directories,
    interpret,
)

TEST_OPTIONS = ['--mca', 'btl_tcp_if_include', 'ib0']
DEFAULT_CONFIGURATION = config.CONFIGURATION
TEST_CONFIGURATION = config.Configuration.create_from_string(f"""
launcher_options: {TEST_OPTIONS}
""")


class LauncherTest(tests.TestCase):

    @classmethod
    def setUpClass(cls):
        tests.TestCase.setUpClass()
        test_module_name = 'assets.test_launcher_module'
        import_module(name=test_module_name)

        for script_name in sys.modules[test_module_name].SCRIPT_NAMES:
            launchers.LAUNCHERS.update({script_name: test_module_name})

    def test_no_launcher(self):
        """
        Check a command is not interpreted as a launcher
        """
        text = 'ls -alh /dev'
        launcher, command = interpret(split(text))

        self.assertFalse(launcher)
        self.assertEqual(command, split(text))

    def test_option(self):
        """
        Check option support
        """
        launcher_txt = 'mylauncher -a -b test'
        command_txt = 'command'
        cmd_line = split(launcher_txt) + split(command_txt)
        launcher, command = interpret(cmd_line)

        self.assertEqual(launcher, split(launcher_txt))
        self.assertEqual(command, split(command_txt))

    #Obselete. We pass user args directly to launchers
    '''
    def test_unsupported_option(self):
        """
        Check unsupported option behaviour
        """
        launcher_txt = 'mylauncher -z 4'
        command_txt = 'command'
        cmd_line = split(launcher_txt) + split(command_txt)
        launcher, command = interpret(cmd_line)

        self.assertNotEqual(launcher, split(launcher_txt))
        self.assertNotEqual(command, split(command_txt))
    '''
    
    def test_equalled_option(self):
        """
        Check support for --x=y style options
        """
        launcher_txt = "mylauncher --option1=a --option2=a:b:c --option3='something or another'"
        command_txt = 'command'
        cmd_line = split(launcher_txt) + split(command_txt)
        launcher, command = interpret(cmd_line)

        self.assertEqual(launcher, split(launcher_txt))
        self.assertEqual(command, split(command_txt))

    def test_get_reserved_directories(self):
        """
        Check access of metadata
        """
        dirs = get_reserved_directories(['mylauncher'])

        self.assertEqual([Path('/reserved')], dirs)

    def test_dashes(self):
        """
        Check dashes separate commands properly
        """
        launcher_txt = 'mylauncher -n 2'
        command_txt = 'hostname'
        cmd_line = [*split(launcher_txt), '--', *split(command_txt)]
        launcher, command = interpret(cmd_line)

        self.assertEqual(launcher, split(launcher_txt))
        self.assertEqual(command, split(command_txt))

        launcher_txt = 'unsupported -z 4'
        command_txt = 'hostname'
        cmd_line = [*split(launcher_txt), '--', *split(command_txt)]
        launcher, command = interpret(cmd_line)

        self.assertEqual(launcher, split(launcher_txt))
        self.assertEqual(command, split(command_txt))

    def test_E_flag_split(self):
        """
        Check complex argument support
        """
        text = "mpirun -E 'two words' command"
        launcher, command = interpret(split(text))

        self.assertEqual(launcher, ['mpirun', '-E', 'two words'])
        self.assertEqual(command, ['command'])

    def test_filter_arguments(self):
        """Check argument filtering support"""
        parser = Parser({'-a': 1, '-c': 2})
        command_line = "-a test -b 1 2 3 -c ofi btl -d host1:2,host2:2".split(
            ' ')

        valid, foreign = filter_arguments(parser, command_line)

        self.assertEqual(valid, "-a test -c ofi btl".split())
        self.assertEqual(foreign, "-b 1 2 3 -d host1:2,host2:2".split())

    def test_concat_arguments(self):
        """
        Check concatenated argument support
        """
        text = "srun -n4 -pgpu -Acourses01-gpu command"
        launcher, command = interpret(split(text))

        self.assertEqual(launcher, ['srun', '-n4', '-pgpu', '-Acourses01-gpu'])
        self.assertEqual(command, ['command'])

    def test_additional_options_config(self):
        """Check configuration options get added to the launcher command"""
        command = ['mpirun', '-np', '4', './foo', '--bar']

        launcher, application = interpret(command)
        self.assertEqual(launcher, ['mpirun', '-np', '4'])
        self.assertEqual(application, ['./foo', '--bar'])

        config.update_configuration(TEST_CONFIGURATION)
        launcher, application = interpret(command)
        self.assertEqual(launcher, ['mpirun', '-np', '4', *TEST_OPTIONS])
        self.assertEqual(application, ['./foo', '--bar'])

        config.update_configuration(DEFAULT_CONFIGURATION)
        launcher, application = interpret(command)
        self.assertEqual(launcher, ['mpirun', '-np', '4'])
        self.assertEqual(application, ['./foo', '--bar'])

    def test_additional_options_environment(self):
        """Check environment options get added to the launcher command"""
        command = ['mpirun', '-np', '4', './foo', '--bar']

        launcher, application = interpret(command)
        self.assertEqual(launcher, ['mpirun', '-np', '4'])
        self.assertEqual(application, ['./foo', '--bar'])

        environ['E4S_CL_LAUNCHER_OPTIONS'] = " ".join(TEST_OPTIONS)
        launcher, application = interpret(command)
        self.assertEqual(launcher, ['mpirun', '-np', '4', *TEST_OPTIONS])
        self.assertEqual(application, ['./foo', '--bar'])

        del environ['E4S_CL_LAUNCHER_OPTIONS']
        launcher, application = interpret(command)
        self.assertEqual(launcher, ['mpirun', '-np', '4'])
        self.assertEqual(application, ['./foo', '--bar'])

    def test_additional_options_priority(self):
        """Check environment options get precedence over configuration"""
        command = ['mpirun', '-np', '4', './foo', '--bar']

        launcher, application = interpret(command)
        self.assertEqual(launcher, ['mpirun', '-np', '4'])
        self.assertEqual(application, ['./foo', '--bar'])

        environ['E4S_CL_LAUNCHER_OPTIONS'] = " ".join([*TEST_OPTIONS, '--env'])
        config.update_configuration(TEST_CONFIGURATION)
        launcher, application = interpret(command)
        self.assertEqual(launcher,
                         ['mpirun', '-np', '4', *TEST_OPTIONS, '--env'])
        self.assertEqual(application, ['./foo', '--bar'])

        del environ['E4S_CL_LAUNCHER_OPTIONS']
        launcher, application = interpret(command)
        self.assertEqual(launcher, ['mpirun', '-np', '4', *TEST_OPTIONS])
        self.assertEqual(application, ['./foo', '--bar'])

        config.update_configuration(DEFAULT_CONFIGURATION)
        launcher, application = interpret(command)
        self.assertEqual(launcher, ['mpirun', '-np', '4'])
        self.assertEqual(application, ['./foo', '--bar'])



from os import getcwd, environ, pathsep
from unittest import skipIf
from pathlib import Path
import tests
from e4s_cl import config
from e4s_cl.util import which
from e4s_cl.cf.containers import (
    BackendUnsupported,
    BoundFile,
    Container,
    FileOptions,
)

CONFIG_EXECUTABLE = tests.ASSETS / 'bin' / 'singularity-conf'
DEFAULT_CONFIGURATION = config.CONFIGURATION
TEST_CONFIGURATION = config.Configuration.create_from_string(f"""
backends:
  singularity:
    executable: '{CONFIG_EXECUTABLE}'
    options: ['--nocolor', '-s']
    exec_options: ['--hostname', 'XxmycoolcontainerxX']
""")


class ContainerTestSingularity(tests.TestCase):

    def test_create(self):
        container = Container(name='singularity', image='test')
        self.assertFalse(type(container) == Container)
        self.assertTrue(isinstance(container, Container))

    def test_run_image(self):
        container = Container(name='singularity', image='imagenametest')
        command = ['']
        container_cmd = container._prepare(command)
        self.assertIn('imagenametest', ' '.join(map(str, container_cmd)))

    def test_run_pwd(self):
        container = Container(name='singularity')
        command = ['']
        container_cmd = container._prepare(command)
        pwd = getcwd()
        self.assertIn(pwd, ' '.join(map(str, container_cmd)))

    def test_run_mpirun(self):
        container = Container(name='singularity', image='dummyimagename')
        command = ['mpirun -n 2 ls']
        container_cmd = container._prepare(command)
        self.assertIn(command[0], ' '.join(map(str, container_cmd)))

    def test_bind_file(self):
        container = Container(name='singularity')

        target = Path('/tmp')
        dest = Path('/etc')
        contained_dest = Path("/etc/skel")
        option = FileOptions.READ_WRITE

        container.bind_file(target)
        self.assertIn(BoundFile(target, target, FileOptions.READ_ONLY),
                      list(container.bound))

        container._Container__bound_files = {}

        container.bind_file(target, dest=dest)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_ONLY),
                      list(container.bound))

        container._Container__bound_files = {}

        container.bind_file(target, dest=dest)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_ONLY),
                      list(container.bound))

        container.bind_file(target, dest=dest, option=option)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_WRITE),
                      list(container.bound))

        container._Container__bound_files = {}

        container.bind_file(target, dest=contained_dest)
        self.assertIn(BoundFile(target, contained_dest, FileOptions.READ_ONLY),
                      list(container.bound))

        container.bind_file(target, dest=dest, option=option)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_WRITE),
                      list(container.bound))

    def test_bind_relative(self):
        container = Container(name='singularity')

        target = Path('/tmp/../proc/meminfo')

        ref = Path('/tmp')
        file_ex = Path('/proc/meminfo')
        home = Path.home()
        paths = {ref, file_ex, home}

        container.bind_file(target)
        files = set(map(lambda x: x.origin, container.bound))

        for item in paths:
            self.assertIn(item, files)

    def test_additional_options_config(self):
        container = Container(name='singularity')
        command = ['']

        singularity_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, singularity_command)

        config.update_configuration(TEST_CONFIGURATION)
        singularity_command = container._prepare(command)
        self.assertContainsInOrder([
            '--nocolor',
            '-s',
            'exec',
            '--hostname',
            'XxmycoolcontainerxX',
        ], singularity_command)

        config.update_configuration(DEFAULT_CONFIGURATION)
        singularity_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, singularity_command)

    def test_additional_options_environment(self):
        container = Container(name='singularity')
        command = ['']

        singularity_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, singularity_command)

        environ['E4S_CL_SINGULARITY_OPTIONS'] = "--nocolor -s"
        environ[
            'E4S_CL_SINGULARITY_EXEC_OPTIONS'] = "--hostname XxmycoolcontainerxX"
        singularity_command = container._prepare(command)
        self.assertContainsInOrder([
            '--nocolor',
            '-s',
            'exec',
            '--hostname',
            'XxmycoolcontainerxX',
        ], singularity_command)

        del environ['E4S_CL_SINGULARITY_OPTIONS']
        del environ['E4S_CL_SINGULARITY_EXEC_OPTIONS']
        singularity_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, singularity_command)

    def test_executable(self):
        """Assert the default singularity executable comes from $PATH"""
        container = Container(name='singularity')

        default_path = environ.get('PATH', '')
        environ['PATH'] = f"{tests.ASSETS / 'bin'}{pathsep}{default_path}"

        self.assertEqual(tests.ASSETS / 'bin' / 'singularity',
                         container._executable())

        environ['PATH'] = default_path

    def test_executable_config(self):
        """Assert the singularity executable is read from the configuration"""
        container = Container(name='singularity')

        config.update_configuration(TEST_CONFIGURATION)
        self.assertEqual(tests.ASSETS / 'bin' / 'singularity-conf',
                         container._executable())

        config.update_configuration(DEFAULT_CONFIGURATION)

    def test_executable_env(self):
        """Assert the singularity executable is read from the environment"""
        container = Container(name='singularity')

        environ['E4S_CL_SINGULARITY_EXECUTABLE'] = str(tests.ASSETS / 'bin' /
                                                       'singularity-env')
        self.assertEqual(tests.ASSETS / 'bin' / 'singularity-env',
                         container._executable())

        del environ['E4S_CL_SINGULARITY_EXECUTABLE']

    def test_executable_priority(self):
        """Assert the environment has precedence over config and config over default"""
        container = Container(name='singularity')

        default_path = environ.get('PATH', '')
        environ['PATH'] = f"{tests.ASSETS / 'bin'}{pathsep}{default_path}"
        config.update_configuration(TEST_CONFIGURATION)
        environ['E4S_CL_SINGULARITY_EXECUTABLE'] = str(tests.ASSETS / 'bin' /
                                                       'singularity-env')

        self.assertEqual(tests.ASSETS / 'bin' / 'singularity-env',
                         container._executable())

        del environ['E4S_CL_SINGULARITY_EXECUTABLE']

        self.assertEqual(tests.ASSETS / 'bin' / 'singularity-conf',
                         container._executable())

        config.update_configuration(DEFAULT_CONFIGURATION)

        self.assertEqual(tests.ASSETS / 'bin' / 'singularity',
                         container._executable())

        environ['PATH'] = default_path

        # This fails as the which wrapper holds a cache
        #self.assertNotEqual(tests.ASSETS / 'bin' / 'singularity', container._executable())



"""
Print profile in its internal representation (JSON format).
Used to parse and/or access profiles in an script-accessible way.

If the profile name is omitted, the selected profile will be used.

.. code::

    $ e4s-cl profile dump single
    [
        {
            "name": "single",
            "backend": null,
            "image": null,
            "source": null,
            "files": [],
            "libraries": ["/usr/lib/libmpi.so"]
        }
    ]

You can also dump multiple profiles at once by specifying multiple profiles:

.. code::
    
    $ e4s-cl profile dump MPICH_3.4.2 MPICH_3.3.2
    [
        {
            "name": "MPICH_3.4.2",
            "libraries": ["/usr/lib/libz.so.1", [...], "/usr/lib/libicuuc.so.70"],
            "files": ["/usr/share/hwdata/pci.ids"],
            "source": null,
            "image": null,
            "backend": "None",
            "wi4mpi": null},
        {
            "name": "MPICH_3.3.2",
            "libraries": ["/usr/lib/libm.so.6", [...], "/usr/lib/libc.so.6"],
            "files": ["/usr/share/hwdata/pci.ids"],
            "image": null,
            "source": null,
            "wi4mpi": null,
            "backend": null
        }
    ]
"""

from e4s_cl.cli.cli_view import DumpCommand
from e4s_cl.model.profile import Profile


class ProfileDumpCommand(DumpCommand):

    def __init__(self):
        super().__init__(Profile, __name__)


COMMAND = ProfileDumpCommand()



"""
Ensure the profile select command behaves properly
"""
import tests
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.profile.select import COMMAND


class ProfileSelectTest(tests.TestCase):
    """
    Profiles are erased after each tests
    """
    def setUp(self):
        self.profile_name = 'test_select_profile'

    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()

    def test_select(self):
        Profile.controller().create({'name': self.profile_name})
        self.assertCommandReturnValue(0, COMMAND, [self.profile_name])
        self.assertEqual(Profile.controller().selected().get('name'),
                         self.profile_name)

    def test_select_nonexistent(self):
        self.assertCommandReturnValue(2, COMMAND, ["non-existent"])



"""
Module introducing singularity support
"""

import os
from pathlib import Path
from typing import List
from e4s_cl import logger
from e4s_cl.util import run_subprocess
from e4s_cl.cf.libraries import cache_libraries
from e4s_cl.cf.containers import Container, FileOptions, BackendNotAvailableError

LOGGER = logger.get_logger(__name__)

NAME = 'apptainer'
MIMES = ['.simg', '.sif']

OPTION_STRINGS = {FileOptions.READ_ONLY: 'ro', FileOptions.READ_WRITE: 'rw'}


class ApptainerContainer(Container):
    """
    Class to use when formatting bound files for a singularity execution
    """

    executable_name = 'apptainer'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _working_dir(self):
        return ['--pwd', os.getcwd()]

    def __setup__(self):
        """
        Bind minimal directories stripped from the use of `--contain`
        """
        # Ensure HOME is bound
        self.bind_file(Path.home(), option=FileOptions.READ_WRITE)

        # To use with --contain, but removed as it prevented PMI setup on Theta
        #self.bind_file('/dev', option=FileOptions.READ_WRITE)
        #self.bind_file('/tmp', option=FileOptions.READ_WRITE)

    def _format_bound(self):
        """
        Format a list of files to a compatible bind option of singularity
        """

        def _format():
            for file in self.bound:
                yield f"{file.origin}:{file.destination}:{OPTION_STRINGS[file.option]}"

        self.env.update({"APPTAINER_BIND": ','.join(_format())})

    def _prepare(self, command: List[str], overload: bool = True) -> List[str]:
        """
        Return the command to run in a list of string
        """
        # as of the 12th of July 2022, apptainer v1.0.1 still uses the `.singularity.d` folder
        self.add_ld_library_path("/.singularity.d/libs")
        self.env.update({
            "APPTAINERENV_LD_PRELOAD":
            ":".join(self.ld_preload),
            "APPTAINERENV_LD_LIBRARY_PATH":
            ":".join(self.ld_lib_path),
        })
        self._format_bound()
        nvidia_flag = ['--nv'] if self._has_nvidia() else []

        return [
            *self._additional_options(),
            'exec',
            *self._working_dir(),
            *nvidia_flag,
            *self._additional_options('exec'),
            self.image,
            *command,
        ]

    def bind_env_var(self, key, value):
        self.env.update({f"APPTAINERENV_{key}": value})

    def _has_nvidia(self):
        # Assume that the proper ldconfig call has been run and that nvidia
        # libraries are listed in the cache
        if 'nvidia' not in " ".join(cache_libraries()):
            LOGGER.debug("Disabling Nvidia support: no libraries found")
            return False
        return True

    def run(self, command: List[str], overload: bool = True) -> int:
        executable = self._executable()
        if executable is None:
            raise BackendNotAvailableError(self.__class__.__name__)

        container_cmd = [executable, *self._prepare(command, overload)]

        return run_subprocess(container_cmd, env=self.env)


CLASS = ApptainerContainer



"""
Command stubs, inherited from taucmdr
"""

import re
import json
from texttable import Texttable
from e4s_cl import EXIT_SUCCESS
from e4s_cl import logger, util, cli
from e4s_cl.error import UniqueAttributeError, InternalError, ModelError
from e4s_cl.cf.storage import StorageError
from e4s_cl.cf.storage.levels import SYSTEM_STORAGE, USER_STORAGE
from e4s_cl.cli import arguments
from e4s_cl.cli.command import AbstractCommand


class AbstractCliView(AbstractCommand):
    """A command that works as a `view` for a `controller`.
    
    See http://en.wikipedia.org/wiki/Model-view-controller
    
    Attributes:
        controller (class): The controller class for this view's data.
        model_name (str): The lower-case name of the model.
    """

    # pylint: disable=abstract-method

    def __init__(self,
                 model,
                 module_name,
                 summary_fmt=None,
                 help_page_fmt=None,
                 group=None,
                 include_storage_flag=True):
        self.model = model
        self.model_name = self.model.name.lower()
        format_fields = {'model_name': self.model_name}
        if not summary_fmt:
            summary_fmt = "Create and manage %(model_name)s configurations."
        self.include_storage_flag = include_storage_flag
        super().__init__(module_name,
                         format_fields=format_fields,
                         summary_fmt=summary_fmt,
                         help_page_fmt=help_page_fmt,
                         group=group)


class RootCommand(AbstractCliView):
    """A command with subcommands for actions."""

    def _construct_parser(self):
        usage = f"{self.command} <subcommand> [arguments]"
        epilog = [
            '',
            cli.commands_description(self.module_name), '',
            f"See `{self.command} <subcommand> --help` for more information on a subcommand."
        ]
        parser = arguments.get_parser(prog=self.command,
                                      usage=usage,
                                      description=self.summary,
                                      epilog='\n'.join(epilog))
        parser.add_argument('subcommand',
                            help="See 'subcommands' below",
                            choices=cli.commands_next(self.module_name),
                            metavar='<subcommand>')
        parser.add_argument('options',
                            help="Arguments to be passed to <subcommand>",
                            metavar='[arguments]',
                            nargs=arguments.REMAINDER)
        return parser

    def main(self, argv):
        args = self._parse_args(argv)
        return cli.execute_command([args.subcommand], args.options,
                                   self.module_name)


class CreateCommand(AbstractCliView):
    """Base class for the `create` command of command line views."""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('summary_fmt',
                          "Create %(model_name)s configurations.")
        super().__init__(*args, **kwargs)

    def _construct_parser(self):
        usage = f"{self.command} <{self.model_name}_{self.model.key_attribute}> [arguments]"
        parser = arguments.get_parser_from_model(self.model,
                                                 prog=self.command,
                                                 usage=usage,
                                                 description=self.summary)
        if self.include_storage_flag:
            arguments.add_storage_flag(parser, "create", self.model_name)
        return parser

    def _create_record(self, store, data):
        """Create the model record.
        
        Args:
            store (AbstractStorage): Storage to contain the record.
            data (dict): Record data.
            
        Returns:
            int: :any:`EXIT_SUCCESS` if successful.
        
        Raises:
            UniqueAttributeError: A record with the same unique attribute already exists.
        """
        ctrl = self.model.controller(store)
        key_attr = self.model.key_attribute
        key = data[key_attr]
        try:
            ctrl.create(data)
        except UniqueAttributeError:
            self.parser.error(
                f"A {self.model_name} with {key_attr}='{key}' already exists")
        return EXIT_SUCCESS

    def main(self, argv):
        args = self._parse_args(argv)
        store = arguments.parse_storage_flag(args)[0]
        data = {
            attr: getattr(args, attr)
            for attr in self.model.attributes if hasattr(args, attr)
        }
        return self._create_record(store, data)


class DeleteCommand(AbstractCliView):
    """Base class for the `delete` subcommand of command line views."""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('summary_fmt',
                          "Delete %(model_name)s configurations.")
        super().__init__(*args, **kwargs)

    def _construct_parser(self):
        key_attr = self.model.key_attribute
        usage = f"{self.command} <{self.model_name}_{key_attr}> [arguments]"
        epilog = util.color_text("WARNING: THIS OPERATION IS NOT REVERSABLE!",
                                 'yellow',
                                 attrs=['bold'])
        parser = arguments.get_parser(prog=self.command,
                                      usage=usage,
                                      description=self.summary,
                                      epilog=epilog)
        parser.add_argument(
            key_attr,
            help=
            f"{key_attr.capitalize()} of {self.model_name} configuration to delete",
            nargs="+",
            type=arguments.wildcard_defined_object(self.model, key_attr),
            metavar=f"<{self.model_name}_{key_attr}>")
        if self.include_storage_flag:
            arguments.add_storage_flag(parser, "delete", self.model_name)
        return parser

    def _delete_record(self, store, obj):
        key_attr = self.model.key_attribute
        key = obj.get(key_attr)
        ctrl = self.model.controller(store)
        if not ctrl.exists({key_attr: key}):
            self.parser.error(
                f"No {store.name}-level {self.model_name} with {key_attr}='{key}'."
            )
        ctrl.delete({key_attr: key})
        self.logger.info("Deleted %s '%s'", self.model_name, key)
        return EXIT_SUCCESS

    def main(self, argv):
        args = self._parse_args(argv)
        store = arguments.parse_storage_flag(args)[0]
        objects = getattr(args, self.model.key_attribute)
        for obj in objects:
            for elem in obj:
                self._delete_record(store, elem)
        return EXIT_SUCCESS


class EditCommand(AbstractCliView):
    """Base class for the `edit` subcommand of command line views."""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('summary_fmt',
                          "Modify %(model_name)s configurations.")
        self.include_new_key_flag = kwargs.pop('include_new_key_flag', True)
        super().__init__(*args, **kwargs)

    def _construct_parser(self):
        key_attr = self.model.key_attribute
        usage = f"{self.command} <{self.model_name}_{key_attr}> [arguments]"
        parser = arguments.get_parser_from_model(self.model,
                                                 use_defaults=False,
                                                 prog=self.command,
                                                 usage=usage,
                                                 description=self.summary)
        if self.include_new_key_flag:
            group = parser.add_argument_group(f'{self.model_name} arguments')
            group.add_argument(f"--new-{key_attr}",
                               help=f"change the configuration's {key_attr}",
                               metavar=f"<new_{key_attr}>",
                               dest='new_key',
                               default=arguments.SUPPRESS)
        if self.include_storage_flag:
            arguments.add_storage_flag(parser, "modify", self.model_name)
        return parser

    def _update_record(self, store, data, key):
        ctrl = self.model.controller(store)
        key_attr = self.model.key_attribute
        if not ctrl.exists({key_attr: key}):
            self.parser.error(
                f"No {ctrl.storage.name}-level {self.model_name} with {key_attr}='{key}'."
            )
        ctrl.update(data, {key_attr: key})
        self.logger.info("Updated %s '%s'", self.model_name, key)
        return EXIT_SUCCESS

    def main(self, argv):
        args = self._parse_args(argv)
        store = arguments.parse_storage_flag(args)[0]
        data = {
            attr: getattr(args, attr)
            for attr in self.model.attributes if hasattr(args, attr)
        }
        key_attr = self.model.key_attribute
        try:
            data[key_attr] = args.new_key
        except AttributeError:
            pass
        key = getattr(args, key_attr)
        return self._update_record(store, data, key)


class ListCommand(AbstractCliView):
    """Base class for the `list` subcommand of command line views."""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('summary_fmt',
                          "Show %(model_name)s configuration data.")
        default_style = kwargs.pop('default_style', 'dashboard')
        dashboard_columns = kwargs.pop('dashboard_columns', None)
        title_fmt = kwargs.pop(
            'title_fmt', "%(model_name)s Configurations (%(storage_path)s)")
        super().__init__(*args, **kwargs)
        key_attr = self.model.key_attribute
        self._format_fields = {
            'command': self.command,
            'model_name': self.model_name,
            'key_attr': key_attr
        }
        self.default_style = default_style
        self.dashboard_columns = dashboard_columns or [{
            'header':
            key_attr.title(),
            'value':
            key_attr
        }]
        self.title_fmt = title_fmt

    def short_format(self, models):
        """Format modeled records in short format.

        Args:
            models: Modeled records to format.

        Returns:
            str: Record data in short format.
        """
        return [str(model[self.model.key_attribute]) for model in models]

    def dashboard_format(self, records):
        """Format modeled records in dashboard format.

        Args:
            records: Modeled records to format.
 
        Returns:
            str: Record data in dashboard format.
        """
        title = util.hline(
            self.title_fmt % {
                'model_name': records[0].name.capitalize(),
                'storage_path': records[0].storage
            }, 'cyan')
        header_row = [col['header'] for col in self.dashboard_columns]
        rows = [header_row]
        for record in records:
            row = []
            for col in self.dashboard_columns:
                if 'value' in col:
                    try:
                        cell = record[col['value']]
                    except KeyError:
                        cell = 'N/A'
                elif 'yesno' in col:
                    cell = 'Yes' if record.get(col['yesno'], False) else 'No'
                elif 'function' in col:
                    cell = col['function'](record)
                else:
                    raise InternalError(f"Invalid column definition: {col}")
                row.append(cell)
            rows.append(row)
        table = Texttable(logger.LINE_WIDTH)
        table.set_cols_align(
            [col.get('align', 'c') for col in self.dashboard_columns])
        table.add_rows(rows)
        return [title, table.draw(), '']

    def _format_long_item(self, key, val):
        attrs = self.model.attributes[key]
        if 'collection' in attrs:
            foreign_model = attrs['collection']
            foreign_keys = []
            for foreign_record in val:
                try:
                    foreign_keys.append(
                        str(foreign_record[foreign_model.key_attribute]))
                except (AttributeError, ModelError):
                    foreign_keys.append(str(foreign_record))
            val = ', '.join(foreign_keys)
        elif 'model' in attrs:
            foreign_model = attrs['model']
            try:
                val = str(val[foreign_model.key_attribute])
            except (AttributeError, ModelError):
                val = str(val)
        elif 'type' in attrs:
            if attrs['type'] == 'boolean':
                val = str(bool(val))
            elif attrs['type'] == 'array':
                val = ', '.join(str(x) for x in val)
            elif attrs['type'] != 'string':
                val = str(val)
        else:
            raise InternalError(f"Attribute has no type: {attrs}, {val}")
        description = attrs.get('description', 'No description')
        description = description[0].upper() + description[1:] + "."
        return [key, val, description]

    def long_format(self, records):
        """Format records in long format.
        
        Args:
            records: Controlled records to format.
        
        Returns:
            str: Record data in long format.
        """
        title = util.hline(
            self.title_fmt % {
                'model_name': records[0].name.capitalize(),
                'storage_path': records[0].storage
            }, 'cyan')
        retval = [title]
        for record in records:
            rows = [['Attribute', 'Value', 'Description']]
            for key, val in sorted(record.items()):
                if key != self.model.key_attribute:
                    rows.append(self._format_long_item(key, val))
            table = Texttable(logger.LINE_WIDTH)
            table.set_cols_align(['r', 'c', 'l'])
            table.set_deco(Texttable.HEADER | Texttable.VLINES)
            table.add_rows(rows)
            retval.append(util.hline(record[self.model.key_attribute], 'cyan'))
            retval.extend([table.draw(), ''])
        return retval

    def _construct_parser(self):
        key_str = self.model_name + '_' + self.model.key_attribute
        usage_head = (
            f"{self.command} [{key_str}] [{key_str}] ... [arguments]")
        parser = arguments.get_parser(prog=self.command,
                                      usage=usage_head,
                                      description=self.summary)
        parser.add_argument(
            'keys',
            help=("Show only %(model_name)ss with the given %(key_attr)ss" %
                  self._format_fields),
            metavar=key_str,
            nargs='*',
            default=arguments.SUPPRESS)
        style_dest = 'style'
        style_group = parser.add_mutually_exclusive_group()
        style_group.add_argument('-s',
                                 '--short',
                                 help="show minimal %(model_name)s data" %
                                 self._format_fields,
                                 const='short',
                                 action='store_const',
                                 dest=style_dest,
                                 default=arguments.SUPPRESS)
        style_group.add_argument(
            '-d',
            '--dashboard',
            help="show %(model_name)s data in a fancy dasboard" %
            self._format_fields,
            const='dashboard',
            action='store_const',
            dest=style_dest,
            default=arguments.SUPPRESS)
        style_group.add_argument(
            '-l',
            '--long',
            help="show all %(model_name)s data in a list" %
            self._format_fields,
            const='long',
            action='store_const',
            dest=style_dest,
            default=arguments.SUPPRESS)
        if self.include_storage_flag:
            arguments.add_storage_flag(parser,
                                       "show",
                                       self.model_name,
                                       plural=True,
                                       exclusive=False)
        return parser

    def _list_records(self, storage_levels, keys, style):
        """Shows record data via `print`.
        
        Args:
            storage_levels (list): Storage levels to query, e.g. ['user', 'profile']
            keys (list): Keys to match to :any:`self.key_attr`.
            style (str): Style in which to format records.
            
        Returns:
            int: :any:`EXIT_SUCCESS` if successful.
        """
        user_ctl = self.model.controller(USER_STORAGE)
        system_ctl = self.model.controller(SYSTEM_STORAGE)

        system = SYSTEM_STORAGE.name in storage_levels
        user = USER_STORAGE.name in storage_levels

        parts = []
        if system:
            parts.extend(self._format_records(system_ctl, style, keys))
        if user:
            parts.extend(self._format_records(user_ctl, style, keys))
        if style == 'dashboard':
            # Show record counts (not the records themselves) for other storage levels
            if not system:
                parts.extend(self._count_records(system_ctl))
            if not user:
                parts.extend(self._count_records(user_ctl))
        print("\n".join(parts))
        return EXIT_SUCCESS

    def main(self, argv):
        args = self._parse_args(argv)
        keys = getattr(args, 'keys', None)
        style = getattr(args, 'style', None) or self.default_style
        storage_levels = [l.name for l in arguments.parse_storage_flag(args)]
        return self._list_records(storage_levels, keys, style)

    def _retrieve_records(self, ctrl, keys):
        """Retrieve modeled data from the controller.
        
        Args:
            ctrl (Controller): Controller for the data model.
            keys (list): Keys to match to :any:`self.key_attr`.
            
        Returns:
            list: Model records.
        """
        if not keys:
            records = ctrl.all()
        else:
            key_attr = self.model.key_attribute
            matches = [
                ctrl.match(key_attr, regex=f"^{re.escape(key)}.*")
                for key in keys
            ]

            for i, record in enumerate(matches):
                if not record:
                    self.parser.error(
                        f"No {self.model_name} with {key_attr} matching '{keys[i]}'"
                    )

            records = list(set(util.flatten(matches)))

        return records

    def _format_records(self, ctrl, style, keys=None):
        """Format records in a given style.
        
        Retrieves records for controller `ctrl` and formats them.
        
        Args:
            ctrl (Controller): Controller for the data model.
            style (str): Style in which to format records.        
            keys (list): Keys to match to :any:`self.key_attr`.
            
        Returns:
            list: Record data as formatted strings.
        """
        try:
            records = self._retrieve_records(ctrl, keys)
        except StorageError:
            records = []
        if not records:
            parts = [f"No {self.model_name}s."]
        else:
            formatter = getattr(self, style + '_format')
            parts = formatter(records)
        return parts

    def _count_records(self, ctrl):
        """Print a record count to stdout.
        
        Args:
            controller (Controller): Controller for the data model.
        """
        level = ctrl.storage.name
        try:
            count = ctrl.count()
        except StorageError:
            count = 0
        fields = dict(self._format_fields,
                      count=count,
                      level=level,
                      level_flag=arguments.STORAGE_LEVEL_FLAG)
        if count == 1:
            return [
                "There is 1 %(level)s %(model_name)s."
                " Type `%(command)s -%(level_flag)s %(level)s` to list it." %
                fields
            ]
        if count > 1:
            return [
                "There are %(count)d %(level)s %(model_name)ss."
                " Type `%(command)s -%(level_flag)s %(level)s` to list them." %
                fields
            ]
        #return ["There are no %(level)s %(model_name)ss." % fields]
        return []


class DumpCommand(AbstractCliView):
    """
    Dump json data from the database
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('summary_fmt',
                          "Dump %(model_name)s configuration data.")
        super().__init__(*args, **kwargs)
        key_attr = self.model.key_attribute
        self._format_fields = {
            'command': self.command,
            'model_name': self.model_name,
            'key_attr': key_attr
        }

    def _construct_parser(self):
        key_str = self.model_name + '_' + self.model.key_attribute
        usage_head = (
            "%(command)s [%(key_str)s] [%(key_str)s] ... [arguments]" % {
                'command': self.command,
                'key_str': key_str
            })
        parser = arguments.get_parser(prog=self.command,
                                      usage=usage_head,
                                      description=self.summary)
        parser.add_argument(
            'keys',
            help=(
                "Dump only the %(model_name)ss with the given %(key_attr)ss" %
                self._format_fields),
            metavar=key_str,
            nargs='*',
            default=arguments.SUPPRESS)
        if self.include_storage_flag:
            arguments.add_storage_flag(parser,
                                       "show",
                                       self.model_name,
                                       plural=True,
                                       exclusive=False)
        return parser

    def main(self, argv):
        args = self._parse_args(argv)
        keys = getattr(args, 'keys', None)
        storage_levels = [l.name for l in arguments.parse_storage_flag(args)]
        return self._dump_records(storage_levels, keys)

    def _retrieve_records(self, ctrl, keys):
        """Retrieve modeled data from the controller.
        
        Args:
            ctrl (Controller): Controller for the data model.
            keys (list): Keys to match to :any:`self.key_attr`.
            
        Returns:
            list: Model records.
        """
        if not keys:
            records = ctrl.all()
        else:
            key_attr = self.model.key_attribute
            if len(keys) == 1:
                records = ctrl.search({key_attr: keys[0]})
                if not records:
                    self.parser.error(
                        f"No {self.model_name} with {key_attr} '{keys[0]}'")
            else:
                records = ctrl.search([{key_attr: key} for key in keys])
                for i, record in enumerate(records):
                    if not record:
                        self.parser.error(
                            f"No {self.model_name} with {key_attr} '{keys[i]}'"
                        )
        return records

    def _dump_records(self, storage_levels, keys):
        """Shows record data via `print`.
        
        Args:
            storage_levels (list): Storage levels to query, e.g. ['user', 'profile']
            keys (list): Keys to match to :any:`self.key_attr`.
            
        Returns:
            int: :any:`EXIT_SUCCESS` if successful.
        """
        user_ctl = self.model.controller(USER_STORAGE)
        system_ctl = self.model.controller(SYSTEM_STORAGE)

        system = SYSTEM_STORAGE.name in storage_levels
        user = USER_STORAGE.name in storage_levels

        parts = []
        if system:
            try:
                parts += self._retrieve_records(system_ctl, keys)
            except StorageError:
                pass

        if user:
            try:
                parts += self._retrieve_records(user_ctl, keys)
            except StorageError:
                pass

        print(json.dumps(parts))

        return EXIT_SUCCESS


class ShowCommand(AbstractCliView):
    """
    Print details on an object
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('summary_fmt',
                          "Show %(model_name)s configuration data.")
        super().__init__(*args, **kwargs)
        self._format_fields = {
            'command': self.command,
            'model_name': self.model_name,
            'key_attr': self.model.key_attribute
        }

    def _construct_parser(self):
        usage = ("%(command)s [arguments] <%(model_name)s_%(key_attr)s>" %
                 self._format_fields)
        parser = arguments.get_model_identifier(self.model,
                                                prog=self.command,
                                                usage=usage,
                                                description=self.summary)
        return parser

    def main(self, argv):
        return EXIT_SUCCESS


class CopyCommand(CreateCommand):
    """Base class for the `copy` subcommand of command line views."""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('summary_fmt', "Copy %(model_name)s.")
        super().__init__(*args, **kwargs)

    def _construct_parser(self):
        key_attr = self.model.key_attribute
        usage = f"{self.command} <{self.model_name}_{key_attr}> <copy_{key_attr}> [arguments]"
        parser = arguments.get_model_identifier(self.model,
                                                prog=self.command,
                                                usage=usage,
                                                description=self.summary)
        group = parser.add_argument_group(f'{self.model_name} arguments')
        group.add_argument(
            f'copy_{key_attr}',
            help=f"new {self.model_name} configuration's {key_attr}",
            metavar=f'<copy_{key_attr}>',
            default=arguments.SUPPRESS)
        if self.include_storage_flag:
            arguments.add_storage_flag(parser, "copy", self.model_name)
        return parser

    def _copy_record(self, store, updates, key):
        ctrl = self.model.controller(store)
        key_attr = self.model.key_attribute
        matching = ctrl.search({key_attr: key})
        found = []
        if not matching:
            self.parser.error(
                f"No {ctrl.storage.name}-level {self.model_name} with {key_attr}='{key}'."
            )
        elif len(matching) > 1:
            raise InternalError(
                f"More than one {ctrl.storage.name}-level {self.model_name} with {key_attr}='{key}' exists!"
            )
        else:
            found = matching[0]
        data = dict(found)
        data.update(updates)
        return self._create_record(store, data)

    def main(self, argv):
        args = self._parse_args(argv)
        store = arguments.parse_storage_flag(args)[0]
        _object = getattr(args, self.model.name.lower())
        data = {
            attr: getattr(args, attr)
            for attr in self.model.attributes if hasattr(args, attr)
        }
        key_attr = self.model.key_attribute
        try:
            data[key_attr] = getattr(args, f'copy_{key_attr}')
        except AttributeError:
            pass
        key = _object.get(key_attr)
        return self._copy_record(store, data, key)



"""Utility functions.

Handles system manipulation and status tasks, e.g. subprocess management or file creation.
"""

import os
import re
import sys
import subprocess
from subprocess import DEVNULL, STDOUT
import pkgutil
from pathlib import Path
from hashlib import sha256
import json
import unicodedata
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)
from functools import lru_cache, reduce
from shutil import which as sh_which
from collections import deque
from tarfile import TarFile
from e4s_cl import (
    E4S_CL_ENV_PREFIX,
    logger,
)
from e4s_cl.variables import ParentStatus
from e4s_cl.error import InternalError

try:
    import termcolor
    COLOR_OUTPUT = True
except ModuleNotFoundError:
    COLOR_OUTPUT = False

LOGGER = logger.get_logger(__name__)

# Suppress debugging messages in optimized code
if __debug__:
    _heavy_debug = LOGGER.debug  # pylint: disable=invalid-name
else:

    def _heavy_debug(*args, **kwargs):
        # pylint: disable=unused-argument
        pass


_PY_SUFFEXES = ('.py', '.pyo', '.pyc')

# Don't make this a raw string!  \033 is unicode for '\x1b'.
_COLOR_CONTROL_RE = re.compile('\033\\[([0-9]|3[0-8]|4[0-8])m')


def mkdirp(path: Path) -> bool:
    """Creates a directory and all its parents.
    
    Works just like ``mkdir -p``.
    
    Args:
        path: Path to create.
    """

    if not isinstance(path, Path):
        path = Path(path)

    try:
        path.mkdir(parents=True, exist_ok=True)
    except PermissionError as err:
        LOGGER.debug("Failed to create directory %s: %s", path.as_posix(),
                     str(err))
        return False
    except FileExistsError as err:
        LOGGER.debug("File %s exists and is not a directory: %s",
                     path.as_posix(), str(err))
        return False

    return True


@lru_cache()
def which(*args, **kwargs):
    return sh_which(*args, **kwargs)


def get_env(var: str) -> Optional[str]:
    """Check the environment for a variable. Automatically adds a prefix"""
    marker = f"{E4S_CL_ENV_PREFIX}_{var.upper()}"

    return os.environ.get(
        marker,
        None,
    )


def path_accessible(path: Path, mode: str = 'r') -> bool:
    """Check if a file or directory exists and is accessible.
    
    Files are checked by attempting to open them with the given mode.
    Directories are checked by testing their access bits only, which may fail for 
    some filesystems which may have permissions semantics beyond the usual POSIX 
    permission-bit model. We'll fix this if it becomes a problem. 
    
    Args:
        path (str): Path to file or directory to check.
        mode (str): File access mode to test, e.g. 'r' or 'rw'
    
    Returns:
        True if the file exists and can be opened in the specified mode, False otherwise.
    """
    if isinstance(path, str):
        path = Path(path)

    modes = {'r': os.R_OK, 'w': os.W_OK, 'x': os.X_OK}

    if not mode:
        raise InternalError(f"Unsupported value for mode: '{mode}'")
    for element in mode:
        if element not in modes:
            raise InternalError(f"Unsupported value for mode: '{element}'")

    modebits = 0
    for char in mode:
        modebits |= modes[char]
    return os.access(path.as_posix(), os.F_OK) and os.access(
        path.as_posix(), modebits)


def run_subprocess(cmd, cwd=None, env=None, discard_output=False) -> int:
    """
    cmd: list[str],
    env: Optional[dict]
    Run a subprocess, tailored for end subrocesses
    """

    subproc_env = os.environ
    stdout, stderr = sys.stdout, subprocess.PIPE
    if discard_output:
        stdout, stderr = DEVNULL, STDOUT
    if env:
        for key, val in env.items():
            if val is None:
                subproc_env.pop(key, None)
                _heavy_debug("unset %s", key)
            else:
                subproc_env[key] = val
                _heavy_debug("%s=%s", key, val)

    # Store the N last lines of error in a fast container
    buffer = deque(maxlen=100)

    with subprocess.Popen(cmd,
                          cwd=cwd,
                          env=subproc_env,
                          stdout=stdout,
                          stderr=stderr,
                          close_fds=False,
                          universal_newlines=True,
                          bufsize=1) as proc:
        # Save the PID for later use
        pid = proc.pid
        # Setup a logger dedicated to this subprocess
        process_logger = logger.setup_process_logger(f"process.{pid}")
        if not discard_output:
            with proc.stderr:
                # Log the errors in a log file
                for line in proc.stderr.readlines():
                    process_logger.error(line[:-1])
                    sys.stderr.write(line)
                    sys.stderr.flush()
                    buffer.append(line)
        returncode = proc.wait()

    # In case of error, output information
    if returncode:
        LOGGER.error("Process %d failed with code %d", pid, returncode)
        for line in buffer:
            LOGGER.error(line)
        if process_logger.handlers:
            log_file = getattr(process_logger.handlers[0], 'baseFilename',
                               None)
            if log_file:
                LOGGER.error("See %s for details.", log_file)
    else:
        LOGGER.debug("Process %d returned %d", pid, returncode)

    del process_logger

    return returncode


def run_e4scl_subprocess(cmd, cwd=None, env=None, capture_output=False, timeout=None) -> Tuple[int, str]:
    """
    cmd: list[str],
    env: Optional[dict],
    capture_output: bool
    timeout: Optional[float], timeout in seconds
    Run a subprocess, tailored for recursive e4s-cl processes
    """
    with ParentStatus():
        subproc_env = os.environ
        if env:
            for key, val in env.items():
                if val is None:
                    subproc_env.pop(key, None)
                    _heavy_debug("unset %s", key)
                else:
                    subproc_env[key] = val
                    _heavy_debug("%s=%s", key, val)

        LOGGER.debug("Running with parent status: %s", cmd)
        with subprocess.Popen(
                cmd,
                cwd=cwd,
                env=subproc_env,
                stdout=subprocess.PIPE if capture_output else sys.stdout,
                stderr=sys.stderr,
                close_fds=False,
                universal_newlines=True,
                bufsize=1) as proc:

            try:
                output, _ = proc.communicate(timeout=timeout)
                returncode = proc.returncode
            except subprocess.TimeoutExpired:
                LOGGER.warning("Subprocess timed out after %s seconds: %s", timeout, cmd)
                proc.kill()
                # Drain remaining output to avoid deadlock when process has buffered data
                # communicate() will collect any remaining output and wait for termination
                try:
                    output, _ = proc.communicate(timeout=5)  # Give 5s for cleanup
                except subprocess.TimeoutExpired:
                    # Process still hasn't terminated after kill+communicate, force wait
                    proc.wait()
                    output = ""
                returncode = -1

    if capture_output:
        return returncode, output
    return returncode, ''


def get_command_output(cmd):
    """Return the possibly cached output of a command.
    
    Just :any:`subprocess.check_output` with a cache.
    Subprocess stderr is always sent to subprocess stdout.
    
    Args:
        cmd (list): Command and its command line arguments.

    Raises:
        subprocess.CalledProcessError: return code was non-zero.
        
    Returns:
        str: Subprocess output.
    """
    key = repr(cmd)
    try:
        return get_command_output.cache[key]
    except AttributeError:
        get_command_output.cache = {}
    except KeyError:
        pass
    else:
        _heavy_debug("Using cached output for command: %s", cmd)
    LOGGER.debug("Checking subprocess output: %s", cmd)
    stdout = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    get_command_output.cache[key] = stdout
    _heavy_debug(stdout)
    LOGGER.debug("%s returned 0", cmd)
    return stdout


def page_output(output_string):
    """Pipe string to a pager.

    If PAGER is an environment then use that as pager, otherwise
    use `less`.

    Args:
        output_string (str): String to put output.

    """
    if os.environ.get('__E4S_CL_ENABLE_PAGER__', False):
        pager_cmd = os.environ.get('PAGER', 'less -F -R -S -X -K').split(' ')
        with subprocess.Popen(pager_cmd, stdin=subprocess.PIPE) as proc:
            proc.communicate(bytearray(output_string, 'utf-8'))
    else:
        print(output_string)


def parse_bool(value, additional_true=None, additional_false=None):
    """Parses a value to a boolean value.
    
    If `value` is a string try to interpret it as a bool:
    * ['1', 't', 'y', 'true', 'yes', 'on'] ==> True
    * ['0', 'f', 'n', 'false', 'no', 'off'] ==> False
    Otherwise raise TypeError.
    
    Args:
        value: value to parse to a boolean.
        additional_true (list): optional additional string values that stand for True.
        additional_false (list): optional additional string values that stand for False.
        
    Returns:
        bool: True if  `value` is true, False if `value` is false.
        
    Raises:
        ValueError: `value` does not parse.
    """
    true_values = ['1', 't', 'y', 'true', 'yes', 'on']
    false_values = ['0', 'f', 'n', 'false', 'no', 'off', 'none']
    if additional_true:
        true_values.extend(additional_true)
    if additional_false:
        false_values.extend(additional_false)
    if isinstance(value, str):
        value = value.lower()
        if value in true_values:
            return True
        if value in false_values:
            return False
        raise TypeError
    return bool(value)

def list_directory_files(path: Path):
    """Lists all the files of a directory.

    Args:
        path (Path): path of the directory list the files of.

    Returns:
        A list of paths of the files in the given directory.
    """
    file_paths = []
    for file_path in path.iterdir():
        file_paths.append(file_path)
    return file_paths

def empty_dir(path: Path):
    """Empties a directory.

    Args:
        path (Path): path of the directory to empty.
    """
    if not path.is_dir():
        LOGGER.debug("Can't empty {path} directory at it isn't one.")
    for file_path in path.iterdir():
        try:
            if file_path.is_file() or file_path.is_symlink():
                file_path.unlink()
            elif file_path.is_dir():
                file_path.rmdir()
        except Exception:
            pass

def create_symlink(path: Path, dest: Path):
    """Creates a symlink.

    Args:
        path (Path): file address to make the symlink towards.
        dest (Path): path to make the symlink at.
    """

    dest_dir = dest.parent
    if not dest_dir.exists():
        mkdirp(dest)
    if not dest.exists():
        try:
            dest.symlink_to(path)
        except Exception:
            pass

def hline(title, *args, **kwargs):
    """Build a colorful horizontal rule for console output.
    
    Uses :any:`logger.LINE_WIDTH` to generate a string of '=' characters
    as wide as the terminal.  `title` is included in the string near the
    left of the horizontal line. 
    
    Args:
        title (str): Text to put on the horizontal rule.
        *args: Positional arguments to pass to :any:`termcolor.colored`.
        **kwargs: Keyword arguments to pass to :any:`termcolor.colored`.
    
    Returns:
        str: The horizontal rule.
    """
    text = f"== {title} ==".ljust(logger.LINE_WIDTH, '=')
    return color_text(text, *args, **kwargs)


def color_text(text, *args, **kwargs):
    """Use :any:`termcolor.colored` to colorize text.
    
    Args:
        text (str): Text to colorize.
        *args: Positional arguments to pass to :any:`termcolor.colored`.
        **kwargs: Keyword arguments to pass to :any:`termcolor.colored`.
        
    Returns:
        str: The colorized text.
    """
    if sys.stdout.isatty() and COLOR_OUTPUT:
        return termcolor.colored(text, *args, **kwargs)
    return text


def uncolor_text(text):
    """Remove color control chars from a string.
    
    Args:
        text (str): Text to colorize.
        
    Returns:
        str: The text without control chars.
    """
    return re.sub(_COLOR_CONTROL_RE, '', text)


def walk_packages(path, prefix):
    """Fix :any:`pkgutil.walk_packages` to work with Python zip files.

    Python's default :any:`zipimporter` doesn't provide an `iter_modules` method so
    :any:`pkgutil.walk_packages` silently fails to list modules and packages when
    they are in a zip file.  This implementation works around this.
    """

    def seen(path, dct={}):  # pylint: disable=dangerous-default-value
        if path in dct:
            return True
        dct[path] = True
        return False

    for importer, name, ispkg in _iter_modules(path, prefix):
        yield importer, name, ispkg
        if ispkg:
            __import__(name)
            path = getattr(sys.modules[name], '__path__', None) or []
            path = [p for p in path if not seen(p)]
            yield from walk_packages(path, name + '.')


def _iter_modules(paths, prefix):
    # pylint: disable=no-member
    yielded = {}
    for importer, name, ispkg in pkgutil.iter_modules(path=paths,
                                                      prefix=prefix):
        if name not in yielded:
            yielded[name] = True
            yield importer, name, ispkg


def flatten(nested_list: List[List[Any]]) -> List[Any]:
    """Flatten a nested list."""
    return [item for sublist in nested_list for item in sublist]


def hash256(data: Union[bytes, str]) -> str:
    """
    Create a hash from a string
    """
    grinder = sha256()
    if isinstance(data, str):
        data = data.encode()
    grinder.update(data)
    return grinder.hexdigest()


def _json_serializer(obj):
    """
    JSON add-on that will transform classes into dicts, and sets into special
    objects to be decoded back into sets with `util.JSONDecoder`.
    """
    if getattr(obj, '__dict__', False):
        return {'__type': type(obj).__name__, '__dict': obj.__dict__}
    if isinstance(obj, set):
        return {'__type': 'set', '__list': list(obj)}

    return obj


# Dict of methods to use when decoding e4s-cl json. Keys correspond to values
# of the `__type` field.
JSON_HOOKS = {}


def _json_decoder(obj):
    """
    JSON add-on to decode dicts with embedded data from `util.JSONSerializer`
    """
    if obj.get('__type', False):
        if obj['__type'] == 'set':
            return set(obj['__list'])

        if obj['__type'] in JSON_HOOKS:
            return JSON_HOOKS[obj['__type']](obj['__dict'])

    return obj


def json_dumps(*args, **kwargs):
    """
    json.dumps wrapper
    """
    if kwargs.get('default'):
        raise InternalError("Cannot override default from util.json_dumps")

    kwargs['default'] = _json_serializer

    return json.dumps(*args, **kwargs)


def json_loads(*args, **kwargs):
    """
    json.loads wrapper
    """
    if kwargs.get('object_hook'):
        raise InternalError("Cannot override object_hook from util.json_loads")

    kwargs['object_hook'] = _json_decoder

    return json.loads(*args, **kwargs)


def add_dot(string: str) -> str:
    if string[-1] in ['.', '!', '?']:
        return string
    return string + '.'


def path_contains(lhs: Path, rhs: Path):
    """
    Returns true if rhs is in the tree of which lhs is the root
    pathlib's < operator compares alphabetically, so here we are
    """
    index = len(lhs.parts)
    return lhs.parts[:index] == rhs.parts[:index]


def apply_filters(filters: List[Callable[[Any], bool]],
                  iterable: Iterable) -> Iterable:
    result = iterable

    for func in filters:
        result = filter(func, result)

    return result


def safe_tar(archive: TarFile) -> bool:
    """
    Assert a tarfile does not possess members whose paths are not contained in
    the extracted data, e.g. absolute paths or relative paths that escape
    using '..'

    'archive-v1.2.3/build'  -> safe
    '/etc/config'           -> unsafe
    '../../../weird_file'   -> unsafe
    """

    def child(path: Path) -> bool:
        """Check a single path's status as a child from where it started"""

        def weigh(token: str) -> int:
            """Define weights for path tokens"""
            token_weights = {'.': 0, '..': -1, '/': -4096}
            return token_weights.get(token, 1)

        def contains(depth: int, weight: int) -> int:
            """Assert the depth never dips below 0 -> we never escape"""
            depth += weight

            if depth < 0:
                return -4096
            return depth

        ordered_weights = list(map(weigh, path.parts))
        return reduce(contains, [0, *ordered_weights]) >= 0

    for member in archive.getmembers():
        if not child(Path(member.name)):
            LOGGER.error(
                "Safety concern unpacking %s: file %s attempts to write out of the decompressed directory",
                archive.name, member.name)
            return False
    return True


def prepend_library_path(path: Union[str, Path]):
    """
    Prepend 'path' to the LD_LIBRARY_PATH variable
    """
    env = os.environ.get("LD_LIBRARY_PATH", [])
    os.environ["LD_LIBRARY_PATH"] = os.pathsep.join([str(path), *env])


def sanitize_args(argv: List[str]) -> List[str]:
    """
    Cleans a list of command-line arguments by normalizing Unicode characters
    and collapsing all whitespace to a single space.
    """
    sanitized_argv = []
    for arg in argv:
        if isinstance(arg, str):
            # NFKC normalization 
            normalized_str = unicodedata.normalize('NFKC', arg)
            # Collapse all whitespace and strip ends
            sanitized_arg = ' '.join(normalized_str.split())
            sanitized_argv.append(sanitized_arg)
        else:
            # Preserve non-string arguments
            sanitized_argv.append(arg)

    return sanitized_argv



"""
Defines an abstract class to simplify the use of container technology.
Creating an instance of ``Container`` will return a specific class to
the required backend.

To implement support for a new container backend, create a submodule
that posesses the following attributes:
    - NAME: The name identifying the backend
    - EXECUTABLES: List of executables the backend uses; if not accessible on the
        PATH, the backend will be disabled
    - MIMES: if the backend uses files, extensions used to guess the backend from
        an image
    - CLASS: the class to use. The class' `run` method will be used when launching
        the container: refer to its docstring for details
"""

import re
import sys
import json
from dataclasses import dataclass
from importlib import import_module
from tempfile import TemporaryFile, NamedTemporaryFile
from pathlib import Path
from typing import Union, List, Tuple, Iterable, Optional
from sotools.dl_cache import cache_libraries, get_generator
from e4s_cl.logger import get_logger, debug_mode
from e4s_cl import (
    CONTAINER_BINARY_DIR,
    CONTAINER_DIR,
    CONTAINER_LIBRARY_DIR,
    CONTAINER_SCRIPT,
    EXIT_FAILURE,
    config,
)
from e4s_cl.variables import ParentStatus
from e4s_cl.util import (
    get_env,
    json_loads,
    path_contains,
    run_e4scl_subprocess,
    walk_packages,
    which,
)
from e4s_cl.cf.version import Version
from e4s_cl.error import ConfigurationError

LOGGER = get_logger(__name__)

# List of available modules, accessible by their "executable" or cli tool names
BACKENDS = {'barebones': 'e4s_cl.cf.containers.barebones'}

# List of available, non-debug backends, for help and completion
EXPOSED_BACKENDS = []

# Used to identify backends by the image's suffix
# list of tuples containing (suffix, backend_name)
MIMES = []


# pylint: disable=too-few-public-methods
class FileOptions:
    """
    Abstraction of bound files options
    """
    READ_ONLY = 0
    READ_WRITE = 1


@dataclass(frozen=True)  # Frozen to take advantage of the hashing capacities
class BoundFile:
    """Element of the bound file dictionnary"""
    origin: Path
    destination: Path
    option: int = FileOptions.READ_ONLY


class BackendError(ConfigurationError):
    """Error raised when the requested container tech is not available"""

    def __init__(self, backend_name):
        self.offending = backend_name
        self._message = f"An error has been encountered setting up the container technology backend {backend_name}."
        super().__init__(self._message)

    def handle(self, etype, value, tb):
        LOGGER.critical(self._message)
        return EXIT_FAILURE


class BackendNotAvailableError(BackendError):
    """Error raised when the requested backend is not found on the system"""

    def __init__(self, backend_name):
        super().__init__(backend_name)
        self._message = f"Backend {self.offending} not found. Is the module loaded ?"


class BackendUnsupported(BackendError):
    """Error raised when the requested backend is not supported"""

    def __init__(self, backend_name):
        super().__init__(backend_name)
        pretty = 's are' if len(EXPOSED_BACKENDS) > 1 else ' is'
        self._message = f"""Backend {self.offending} not supported at this time.
The available backend{pretty}: {", ".join(EXPOSED_BACKENDS)}.
Please create a GitHub issue if support is required."""


class AnalysisError(ConfigurationError):
    """Generic error for container analysis failure"""

    def __init__(self, returncode):
        self.code = returncode
        super().__init__(f"Container analysis failed ! ({self.code})")

    def handle(self, etype, value, tb):
        LOGGER.critical("Container analysis failed ! (%d)", self.code)
        return EXIT_FAILURE


def dump(func):
    """
    If verbose, output a description of the container about to be run before
    launching it.

    This needs to be a decorator as it is wraps the run() method implemented
    in every backend module.
    """

    def wrapper(*args, **kwargs):
        self = func.__self__

        LOGGER.debug(str(self))

        return func(*args, **kwargs)

    wrapper.__name__ = func.__name__
    return wrapper


def _contains(container: BoundFile, containee: BoundFile) -> bool:
    """
        Assert the containee bind is 'contained' by container bind. This can happen in multiple ways:
        - The origin containee file is inside the origin container file/directory tree and both are bound in-place
          ORIGIN -> DESTINATION
          container: /usr -> /usr
          containee: /usr/lib/openmpi -> /usr/lib/openmpi
        - The origin containee is a symlink to the origin container and both are bound to the same location

        The following is not covered by contains:
        - A directory is bound and one of its files is bound somewhere else:
          ORIGIN -> DESTINATION
          container: /usr -> /usr
          containee: /usr/lib/libtest.so -> /otherpath/libtest.so
        - Two different files are bound to a single destination
        """

    # Check for an exact match between the container and containee
    exact_match = False
    if containee.destination == container.destination:
        exact_match = (
            container.origin.resolve() == containee.origin.resolve())

    # Check for arborescence delta matches
    try:
        arborescence = containee.origin.relative_to(
            container.origin) == containee.destination.relative_to(
                container.destination)
    except ValueError:
        arborescence = False

    return exact_match or arborescence


def optimize_bind_addition(
        new: BoundFile,
        bound_files: Iterable[BoundFile]) -> Iterable[BoundFile]:
    """
    Adds new to bound_files, if needed. Performs optimizations to prevent
    double binds/superfluous binds, and returns the optimized bind set
    """

    new_binds = set(bound_files)

    # Check if the file to be bound is contained in already bound files/folders
    # If it is, check that the containing files/folders' permissions align with
    # the new file and update them if need be
    target_contained = set(
        filter(lambda bind: _contains(bind, new), bound_files))

    if target_contained:
        # Compute the max permission required by the files containing new. If
        # they allow a lower level of permissions, re-bind them with the
        # necessary permissions
        target_contained_permissions = max(
            map(lambda x: x.option, target_contained))

        if target_contained_permissions < new.option:
            # Re create all the binds
            new_contained = set(
                map(lambda x: BoundFile(x.origin, x.destination, new.option),
                    target_contained))

            # Remove the old binds
            new_binds = new_binds - target_contained

            # Add the ones created above
            new_binds = new_binds | new_contained

        return new_binds

    # Check if the file to be bound is containing already bound files/folders
    # If it is, check that the new file's permissions align with the contained files/folders
    # and then unbind them with a new permission level if need be
    target_containing = set(filter(lambda b: _contains(new, b), bound_files))

    if target_containing:
        # Check the permissions requires by the files contained by new, and
        # update new's permissions accordingly
        target_containing_permissions = max(
            map(lambda x: x.option, target_containing))

        if target_containing_permissions > new.option:
            new = BoundFile(new.origin, new.destination,
                            target_containing_permissions)

    new_binds.add(new)
    return new_binds - target_containing


def _unrelative(string: str) -> Iterable[Path]:
    """
    Returns a list of all the directories referenced by a relative path
    """

    path = Path(string)
    visited = {path, path.resolve()}
    deps = set()

    for i, part in enumerate(path.parts):
        if part == '..':
            visited.add(Path(*path.parts[:i]).resolve())

    for element in visited:
        contained = False
        for path in visited:
            if path != element and path_contains(path, element):
                contained = True

        if not contained:
            deps.add(element)

    return deps


class Container:
    """
    Abstract class that auto-completes depending on the container tech
    """

    # pylint: disable=unused-argument
    def __new__(cls, image=None, name=None):
        """
        Object level creation hijacking: depending on the executable
        argument, the appropriate subclass will be returned.
        """
        module_name = BACKENDS.get(name)
        if module_name:
            module = sys.modules.get(module_name)
        else:
            raise BackendUnsupported(name)

        driver = object.__new__(module.CLASS)

        # If in debugging mode, print out the config before running
        if debug_mode():
            driver.run = dump(driver.run)
        driver.__str__ = dump(driver.__str__)

        return driver

    def __init__(self, image=None, name=None):
        """
        Common class init: this code is run in the actual sub-classes
        """

        # Container image identifier
        self.image = image

        # Container type identifier
        self.name = name

        # User-set parameters
        # Files to bind: set(BoundFile)
        self._bound_files = set()
        self.env = {}  # Environment
        self.ld_preload = []  # Files to put in LD_PRELOAD
        self.ld_lib_path = []  # Directories to put in LD_LIBRARY_PATH

        self.libc_v = Version('0.0.0')
        self.cache = {}

        if hasattr(self, '__setup__'):
            self.__setup__()

    @property
    def script(self):
        return Path(CONTAINER_SCRIPT)

    @property
    def import_dir(self):
        return Path(CONTAINER_DIR)

    @property
    def import_library_dir(self):
        return Path(CONTAINER_LIBRARY_DIR)

    @property
    def import_binary_dir(self):
        return Path(CONTAINER_BINARY_DIR)

    def _executable(self) -> Optional[Path]:
        """
        Inspect the configuration and environment for a path towards the
        container binary to execute.

        If nothing is found, the default executable name is looked up on $PATH.
        """

        def _valid_executable(path: Path) -> bool:
            # Check for executable status ?
            return path.exists() and path.is_file()

        container_type_id = getattr(self, 'name', None)
        if container_type_id is None or not isinstance(container_type_id, str):
            return None

        # Generate the marker from the parameters.
        # If kind is present -> {container_id}_{kind}_options
        # If not             -> {container_id}_options
        marker = f"{container_type_id}_executable"

        def _check(path: str, origin: str, marker: str) -> bool:
            if not path:
                return False

            path = Path(path)
            if _valid_executable(path):
                LOGGER.debug(
                    "%s container executable (from %s): %s",
                    container_type_id,
                    marker,
                    env_option,
                )
                return True

            LOGGER.error(
                "Invalid executable path for %s: %s",
                container_type_id,
                path,
            )

            return False

        # Fetch the options from the environment first
        env_option = get_env(marker)
        if _check(env_option, 'env', marker.upper()):
            return Path(env_option)

        # If the environment is empty, try the configuration
        config_option = getattr(
            config.CONFIGURATION,
            f"backends_{marker}",
            None,
        )

        if _check(config_option, 'config', marker):
            return Path(config_option)

        default_executable = getattr(self.__class__, 'executable_name', None)
        if default_executable is not None:
            path = which(default_executable)

            if path is not None and _valid_executable(Path(path)):
                return Path(path)

        return None

    def _additional_options(self, kind: Optional[str] = None) -> List[str]:
        """
        Inspect the configuration and environment to get a list of additional
        options for the given container.

        The `kind` argument allows to maintain multiple sets of arguments, to
        be used differently, for each container module.

        Options set in the environment have higher priority over options set in
        configuration files.
        """

        container_type_id = getattr(self, 'name', None)
        if container_type_id is None or not isinstance(container_type_id, str):
            return []

        # Generate the marker from the parameters.
        # If kind is present -> {container_id}_{kind}_options
        # If not             -> {container_id}_options
        marker = "_".join(filter(None, [container_type_id, kind, "options"]))

        # Fetch the options from the environment first
        env_options = get_env(marker)
        if env_options:
            LOGGER.debug(
                "%s container additional options (from env %s): %s",
                container_type_id,
                marker.upper(),
                env_options.split(),
            )
            return env_options.split()

        # If the environment is empty, try the configuration
        config_options = getattr(
            config.CONFIGURATION,
            f"backends_{marker}",
            None,
        )

        if config_options:
            LOGGER.debug(
                "%s container additional options (from config %s): %s",
                container_type_id,
                marker,
                config_options,
            )
            return config_options

        return []

    def get_data(self):
        """
        Run analysis commands in the container to get informations about the
        environment inside of it. The results will be used to tailor the
        library import to ensure compatibility of the shared objects.

        The entrypoint passed as an argument may contain external parameters
        (The source script being one of them)

        A library set with data about libraries listed in library_set will
        be returned
        """
        outstream = sys.stdout

        if self.cache:
            return set()

        # Obfuscate stdout to access the output of the below commands
        with TemporaryFile() as buffer:
            # Most likely is the source of the errors observed in https://github.com/E4S-Project/e4s-cl/issues/100
            sys.stdout = buffer

            code = self.run(['cat', '/etc/ld.so.cache'], overload=False)

            if code:
                raise AnalysisError(code)

            sys.stdout.seek(0, 0)
            cache_data = sys.stdout.read()

            # Extract version info from the cache
            glib_version_string = get_generator(cache_data)
            if glib_version_string is not None:
                self.libc_v = Version(glib_version_string)
            else:
                # for older caches, grab the version from the ldconfig binary
                sys.stdout.seek(0, 0)
                sys.stdout.truncate(0)
                code = self.run(['ldconfig', '--version'], overload=False)
                sys.stdout.seek(0, 0)
                glib_version = sys.stdout.read().decode()
                self.libc_v = Version(glib_version)

            LOGGER.debug("Detected container glibc version: %s", self.libc_v)

            # Extract libraries from the cache
            with NamedTemporaryFile('wb', delete=False) as cache_buffer:
                cache_buffer.write(cache_data)
                buffer_name = cache_buffer.name

            self.cache = cache_libraries(buffer_name)

        sys.stdout = outstream

        return set()

    def bind_file(self,
                  path: Union[Path, str],
                  dest: Optional[Path] = None,
                  option: int = FileOptions.READ_ONLY) -> None:
        """
        If there is no destination, handle files with relative paths.
        For instance on summit, some files are required as
        /jsm_pmix/container/../lib/../bin/file
        Although only /jsm_pmix/bin/file is required, not
        having jsm_pmix/container && lib makes it error out.
        unrelative returns a list of all the paths required for such a file
        """
        if not path:
            return

        new_binds = set()
        if not dest:
            for _path in _unrelative(path):
                new_binds.add(BoundFile(_path, _path, option))
        else:
            new_binds.add(BoundFile(Path(path), Path(dest), option))

        for bind in new_binds:
            self._bound_files = optimize_bind_addition(bind, self._bound_files)

    @property
    def bound(self):
        for bound in self._bound_files:
            if bound.origin.exists() and bound.destination.is_absolute():
                yield bound
            else:
                LOGGER.warning(
                    "Attempting to bind non-existing file: %(source)s to %(dest)s",
                    {
                        'source': bound.origin,
                        'dest': bound.destination
                    })

    def bind_env_var(self, key, value):
        self.env.update({key: value})

    def add_ld_preload(self, path):
        if path not in self.ld_preload:
            self.ld_preload.append(path)

    def add_ld_library_path(self, path):
        if path not in self.ld_lib_path:
            self.ld_lib_path.append(path)

    def run(self, command: List[str], overload: bool = True) -> int:
        """
        run a command in a container.

        This method must be implemented in the container module.
        It should take into account the parameters set in the object:
        - The bound files in self.bound;
        - The environment variables self.env;
        - The LD_PRELOAD self.ld_preload;
        - The LD_LIBRARY_PATH self.ld_lib_path
        and set them to be available in the created container.

        If the `overload` flag is set to false, the container is started
        without any of the configuration from the Container object. This
        is used to perform analysis commands in a clean environment.
        """
        raise NotImplementedError(
            f"`run` method not implemented for container module {self.__class__.__name__}"
        )

    def __str__(self):
        out = []
        out.append(f"{self.__class__.__name__} object:")
        if self.image:
            out.append(f"- image: {self.image}")
        bound_files = "\n".join([
            f"\t{v.origin} -> {v.destination} ({v.option})" for v in self.bound
        ])
        out.append(f"- bound:\n{bound_files}")
        if self.env:
            out.append(f"- env: { json.dumps(self.env, indent=2)}")
        if self.ld_preload:
            out.append(
                f"- LD_PRELOAD: {json.dumps(self.ld_preload, indent=2)}")
        if self.ld_lib_path:
            out.append(
                f"- LD_LIBRARY_PATH: {json.dumps(self.ld_lib_path, indent=2)}")
        return '\n'.join(out)


def guess_backend(path):
    suffix = Path(path).suffix

    matches = list(filter(lambda x: x[0] == suffix, MIMES))

    # If we cannot associate a unique backend to a MIME
    if len(matches) != 1:
        return None

    return matches[0][1]


def assert_module(_module) -> bool:
    """
    Assert a module defining a container class is properly structured
    """
    required = ['NAME', 'CLASS']

    for attribute in required:
        if not hasattr(_module, attribute):
            LOGGER.warning(
                "Container module '%s' is missing a required attribute: %s; skipping ...",
                _module.__name__, required)
            return False

    if getattr(_module, 'CLASS') and not getattr(_module.CLASS, 'run'):
        LOGGER.warning(
            "Container module '%s' has an incomplete module class; skipping ...",
            _module.__name__)

    return True


for _, _module_name, _ in walk_packages(__path__, prefix=__name__ + "."):
    try:
        import_module(_module_name)
    except ModuleNotFoundError:
        continue

    _module = sys.modules[_module_name]

    if not assert_module(_module):
        continue

    BACKENDS.update({
        _module.NAME: _module_name,
    })

    if not getattr(_module, 'DEBUG_BACKEND', False):
        EXPOSED_BACKENDS.append(_module.NAME)

    for mimetype in getattr(_module, 'MIMES', []):
        MIMES.append((mimetype, _module.NAME))



"""
This command creates a new profile with the name passed as the first argument.
Options control the data added to the fields of the newly-created profile.

.. warning::

    MPI libraries often dynamically use files and libraries without warning, \
            and the absence of those unlisted files during execution more \
            often than not results in a crash or segmentation fault. Please \
            ensure you acknowledge the result of \
            :ref:`profile detect<profile_detect>` when creating an empty profile.

"""

from e4s_cl.cli import arguments
from e4s_cl.cli.cli_view import CreateCommand
from e4s_cl.model.profile import Profile
from e4s_cl.cf.containers import EXPOSED_BACKENDS


class ProfileCreateCommand(CreateCommand):
    """``profile create`` subcommand."""

    def _construct_parser(self):
        usage = f"{self.command} <profile_name>"
        parser = arguments.get_parser_from_model(self.model,
                                                 prog=self.command,
                                                 usage=usage,
                                                 description=self.summary)
        parser.add_argument(
            '--libraries',
            help="Comma-separated list of libraries to add to this profile",
            metavar='l1.so,l2.so,...',
            type=arguments.posix_path_list,
            default=arguments.SUPPRESS,
            dest='libraries')

        parser.add_argument(
            '--files',
            help="Comma-separated list of files to add to this profile",
            metavar='f1,f2,...',
            type=arguments.posix_path_list,
            default=arguments.SUPPRESS,
            dest='files')

        parser.add_argument(
            '--backend',
            help="Container backend for this profile" +
            f" Available backends are: {', '.join(EXPOSED_BACKENDS)}",
            metavar='technology',
            default=arguments.SUPPRESS,
            dest='backend')

        parser.add_argument('--image',
                            help="Path to a image for this profile",
                            metavar='path/to/image',
                            type=arguments.posix_path,
                            default=arguments.SUPPRESS,
                            dest='image')

        parser.add_argument(
            '--source',
            help="Path to a bash script to source before execution",
            metavar='path/to/script',
            type=arguments.posix_path,
            default=arguments.SUPPRESS,
            dest='source')

        parser.add_argument('--wi4mpi',
                            help="Root of the WI4MPI installation to use",
                            metavar='<path>',
                            type=arguments.posix_path,
                            default=arguments.SUPPRESS)

        return parser


SUMMARY = "Create a profile."

COMMAND = ProfileCreateCommand(Profile, __name__, summary_fmt=SUMMARY)



"""
Identify a MPI vendor and version from an ELF binary
"""

import re
import ctypes
from typing import Optional, Callable, Iterable, List, Set
from dataclasses import dataclass
from pathlib import Path
from e4s_cl import logger
from e4s_cl.model.profile import Profile

LOGGER = logger.get_logger(__name__)


@dataclass(frozen=True)
class MPIIdentifier:
    vendor: str
    version: str

    def __str__(self):
        return f"{self.vendor}@{self.version}".replace(' ', '_')


def _suffix_name(name: str, existing_names: set) -> str:
    """Compute a '-N' suffix for new profiles"""
    # Do not append a suffix for the first unique profile
    if name not in existing_names:
        return name

    # An exact match exists, filter the occurrences of 'name-N' (clones)
    # and return name-max(N)+1
    clones = set(
        filter(
            None,
            map(lambda x: re.match(fr"{re.escape(name)}-(?P<ordinal>\d*)", x),
                existing_names)))

    # Try to list all clones of this profile
    ordinals = []
    for clone in clones:
        try:
            ordinals.append(int(clone.group('ordinal')))
        except ValueError:
            pass

    # If there are no clones, this is the second profile, after the original
    ordinal = 2
    if len(ordinals) != 0:
        ordinal = max(ordinals) + 1

    return f"{name}-{ordinal}"


def strip(function):
    """Strip the output of the given function and catch IndexErrors"""

    def wrapper(version_buffer_str):
        version_str = ''

        try:
            version_str = function(version_buffer_str).strip()
        except IndexError:
            pass

        return version_str

    return wrapper


@strip
def _extract_intel_mpi_version(version_buffer_str: str) -> str:
    """
    Parses the typical Intel MPI library version message, eg:
    Intel(R) MPI Library 2019 Update 6 for Linux* OS
    """
    return version_buffer_str.split("Library", 1)[1].split("for", 1)[0]


@strip
def _extract_open_mpi_version(version_buffer_str: str) -> str:
    """
    Parses the typical OpenMPI library version message, eg:
    Open MPI v4.0.1, package: Open MPI Distribution, ident: 4.0.1, repo rev: v4.0.1, Mar 26, 2019
    """
    return version_buffer_str.split("v", 1)[1].split(",", 1)[0]


@strip
def _extract_spectrum_mpi_version(version_buffer_str: str) -> str:
    """
    Parses the typical Spectrum MPI library version message, eg:
    Open MPI v4.0.1, package: Spectrum MPI Distribution, ident: 4.0.1, repo rev: v4.0.1, Mar 26, 2019
    """
    return version_buffer_str.split("v", 1)[1].split(",", 1)[0]


@strip
def _extract_mpich_version(version_buffer_str: str) -> str:
    """
    Parses the typical MPICH library version message, eg:
    MPICH Version:  3.3b2
    MPICH Release date: Mon Apr  9 17:58:42 CDT 2018
    [...]
    """
    return version_buffer_str.split(":", 1)[1].split("M", 1)[0]


@strip
def _extract_cray_mpich_version(version_buffer_str: str) -> str:
    """
    Parses the typical MPICH library version message, eg:
    MPICH Version:  3.3b2
    MPICH Release date: Mon Apr  9 17:58:42 CDT 2018
    [...]
    """
    return version_buffer_str.split("version", 1)[1].split("(", 1)[0]


@strip
def _extract_mvapich_version(version_buffer_str: str) -> str:
    """
    Parses the typical MVAPICH library version message, eg:
    MVAPICH Version:  3.3b2
    MVAPICH Release date: Mon Apr  9 17:58:42 CDT 2018
    [...]
    """
    return version_buffer_str.split(":", 1)[1].split("M", 1)[0]


VENDOR_VERSION_EXTRACTORS = {
    'Intel(R) MPI': _extract_intel_mpi_version,
    'Open MPI': _extract_open_mpi_version,
    'Spectrum MPI': _extract_spectrum_mpi_version,
    'CRAY MPICH': _extract_cray_mpich_version,
    'MPICH': _extract_mpich_version,
    'MVAPICH': _extract_mvapich_version
}


def _get_mpi_handle(path: Path) -> Optional[Callable]:
    """Get a handle to the MPI_Get_library_version symbol given a path to a
    shared object"""
    if not path.exists():
        return None

    try:
        handle = ctypes.CDLL(path.as_posix())
        return getattr(handle, 'MPI_Get_library_version', None)
    except OSError as err:
        LOGGER.debug("Error loading shared object %s: %s", path.as_posix(),
                     str(err))
        return None


def _get_mpi_library_version(path: Path) -> str:
    """Return the output of the MPI_Get_library_version symbol in the MPI
    binary passed as an argument"""

    if isinstance(path, str):
        path = Path(path)

    # C-compatible buffer to run a C handle with
    version_buffer = ctypes.create_string_buffer(3000)
    length = ctypes.c_int()

    # Get a callable towards the C code
    handle = _get_mpi_handle(path)
    if not handle:
        LOGGER.debug("Extracting MPI_Get_library_version from %s failed",
                     path.as_posix())
        return ''

    # Execute the C code to fill the above buffer
    handle(version_buffer, ctypes.byref(length))

    if length:
        return version_buffer.value.decode("utf-8")[:500]
    return ''


def _get_mpi_vendor_version(path: Path) -> Optional[MPIIdentifier]:
    """Return a tuple of string according to the vendor and version of the MPI
    binary passed as an argument"""
    raw_str = _get_mpi_library_version(path)

    # Check for vendor keywords in the buffer
    filtered_buffer = list(
        filter(lambda x: x in raw_str, VENDOR_VERSION_EXTRACTORS.keys()))

    # Skip this binary if none were found
    if not filtered_buffer:
        return None

    # Sort vendors by size, and save the longest match
    filtered_buffer.sort(key=len, reverse=True)
    vendor_name = filtered_buffer[0]

    # Run the corresponding function on the buffer
    # In case of an error, skip this function
    try:
        version_str = VENDOR_VERSION_EXTRACTORS.get(
            vendor_name, lambda x: 'UNKNOWN_VERSION')(raw_str)
    except IndexError:
        return None

    return MPIIdentifier(vendor_name, version_str)


def detect_mpi(path_list: Iterable[Path]) -> Optional[MPIIdentifier]:
    """Parse the binaries from paths passed as arguments to get a `VENDOR@VERSION` string"""
    # Set of all MPI vendors and versions found in the binaries
    version_data = set(filter(None, map(_get_mpi_vendor_version, path_list)))

    # If one consistent vendor has been found
    if len(version_data) == 1:
        return version_data.pop()
    return None


def filter_mpi_libs(libraries: List[Path]) -> Set[Path]:
    """Return a set of MPI libraries from a list of libraries"""

    def _filter_mpi(path: Path):
        return re.match(r'libmpi.*so.*', path.name)

    return set(filter(_filter_mpi, libraries))


def library_install_dir(libraries: Iterable[Path]) -> Optional[Path]:
    """
    Return the installation directory of a given list of libraries, defined as
    the common path stub containing the 'lib' folder
    """

    def _stub(library: Path) -> Optional[Path]:
        """
        Split a path at the 'lib' directory and return the part before
        """
        path_elements = library.parts
        try:
            lib_index = path_elements.index('lib')
        except ValueError:
            # If not 'lib' directory can be found in the library path, assume
            # a '/x/y/z/lib-like-dir/libmpi.so' structure and return '/x/y/z'
            if len(path_elements) > 2:
                lib_index = len(path_elements) - 2
            else:
                return None
        return Path(*path_elements[:lib_index])

    # Get a set of all the returns of _stub, filtering out the null values
    potential_installs = set(filter(None, map(_stub, libraries)))

    if len(potential_installs) == 1:
        return potential_installs.pop()
    return None


def profile_mpi_name(mpi_libs: Iterable[Path]) -> Optional[str]:
    """
    Analyze the profile with the given eid for MPI libraries and rename it
    according to the vendor/version info in the shared object
    """
    controller = Profile.controller()

    # Run the methods in the libraries to get a version
    mpi_id = detect_mpi(mpi_libs)

    if not mpi_id:
        LOGGER.debug("No symbol found in %s", " ".join(map(str, mpi_libs)))
        return None

    LOGGER.debug("Found identifier %s from profile's MPI libraries", mpi_id)

    # Get all profiles matching the new name
    matches = controller.match('name', regex=f"{re.escape(str(mpi_id))}.*")
    matching_names = set(filter(None, map(lambda x: x.get('name'), matches)))

    # Add a suffix to the name to avoid conflict
    return _suffix_name(str(mpi_id), matching_names)



"""
Ensure the profile show command functions as intended
"""

import tests
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.profile.show import COMMAND as command


class ProfileShowTest(tests.TestCase):
    """
    Partial class, as methods are added manually below
    """

    def tearDown(self):
        self.resetStorage()

    def test_show(self):
        Profile.controller().create({"name": 'test01'})
        self.assertCommandReturnValue(0, command, ['test01'])

    def test_existence(self):
        _, stderr = self.assertNotCommandReturnValue(0, command, ['test01'])
        self.assertIn('profile show [arguments] <profile_name>', stderr)
        self.assertIn('profile show: error:', stderr)

    def test_pattern(self):
        Profile.controller().create({"name": 'test01'})
        Profile.controller().create({"name": 'test02'})
        Profile.controller().create({"name": 'test03'})
        Profile.controller().create({"name": 'otherName01'})
        self.assertNotCommandReturnValue(0, command, ['test'])
        self.assertCommandReturnValue(0, command, ['otherName'])


def wrapper(key, value, test_name):
    """
    Generate tests from a simple pattern to ensure all fields are correctly added
    """

    def generated_test(self):
        Profile.controller().create({'name': 'test01', key: value})
        stdout, _ = self.assertCommandReturnValue(0, command, 'test01')
        if isinstance(value, list):
            for element in value:
                self.assertIn(element, stdout)
        else:
            self.assertIn(value, stdout)

    generated_test.__name__ = f"test_{test_name}"

    return generated_test


_fields = [
    ('image', 'image:id', 'image_value'),
    ('backend', 'singularity', 'backend_value'),
    ('source', '/tmp/test.sh', 'source_value'),
    ('wi4mpi', '/usr/packages/wi4mpi', 'wi4mpi_value'),
    ('libraries', [f"lib{no}.so" for no in range(100)], 'libraries_values'),
    ('files', [f"file_{no}" for no in range(100)], 'files_values'),
]

for arguments in _fields:
    test = wrapper(*arguments)
    setattr(ProfileShowTest, test.__name__, test)



"""
Collection of functions to fetch and install Wi4MPI for a given mpi installation
"""

import os
from shutil import rmtree
import tarfile
import urllib.request
from pathlib import Path
from typing import Optional
from e4s_cl import (
    WI4MPI_DIR,
)
from e4s_cl.util import (
    hash256,
    run_subprocess,
    safe_tar,
)
from e4s_cl.logger import get_logger
from e4s_cl.util import which
from e4s_cl.cf.version import Version
from e4s_cl.cf.wi4mpi import (WI4MPI_METADATA)
from e4s_cl.cf.compiler import (
    CompilerVendor,
    VENDOR_BINARIES,
    available_compilers,
)
from elftools.elf.elffile import ELFFile

LOGGER = get_logger(__name__)

WI4MPI_VERSION = Version('3.6.4')
WI4MPI_RELEASE_URL = f"https://github.com/cea-hpc/wi4mpi/archive/refs/tags/v{WI4MPI_VERSION}.tar.gz"
WI4MPI_RELEASE_SHA256 = "be1732a1aed1e2946873951a344b572f11f2a55cd06c634580a9398b5877e22a"
CPU_COUNT = os.cpu_count()

_WI4MPI_COMPILER_STRINGS = {
    CompilerVendor.GNU: 'GNU',
    CompilerVendor.INTEL: 'INTEL',
    CompilerVendor.LLVM: 'LLVM',
    CompilerVendor.PGI: 'PGI',
    CompilerVendor.ARMCLANG: 'ARMCLANG',
    CompilerVendor.FUJITSU: 'FUJITSU',
}


def _select_compiler() -> Optional[int]:
    """Returns the available compiler"""
    compilers = available_compilers()
    if not compilers:
        return None

    compiler = min(compilers)
    if compiler in _WI4MPI_COMPILER_STRINGS:
        return compiler
    return None


def _fetch_release(destination: Path) -> Optional[Path]:
    """Fetch an url and write it to a temporary file"""
    archive_file = destination / f"wi4mpi-{WI4MPI_VERSION}.tgz"

    if not archive_file.exists():
        try:
            with urllib.request.urlopen(WI4MPI_RELEASE_URL) as request:
                if not request.status == 200:
                    return None

                with open(archive_file, mode='w+b') as buffer:
                    buffer.write(request.read())
        except (ValueError, urllib.error.URLError) as err:
            LOGGER.error("Error fetching URL: %s", err)
            return None

    return archive_file


def _download_wi4mpi(destination: Path) -> Optional[Path]:
    """
    Download and extract the TAR archive from 'url' into the directory 'destination'
    """
    if not destination.exists():
        destination.mkdir()

    try:
        archive = _fetch_release(destination)
        if archive is None or not tarfile.is_tarfile(archive):
            LOGGER.error("Downloaded file is not an archive; aborting")
            return None

        with open(archive, 'rb') as data:
            checksum = hash256(data.read())
            if checksum != WI4MPI_RELEASE_SHA256:
                LOGGER.error(
                    "Downloaded file does not match checksum; aborting")
                return None

        with tarfile.open(archive) as data:
            if not safe_tar(data):
                LOGGER.error("Unsafe paths detected in archive; aborting")
                return None

            release_root_dir = min(data.getnames())

            if not (destination / release_root_dir).exists():
                data.extractall(destination)
    except PermissionError as err:
        LOGGER.error("Failed to download Wi4MPI sources: %s", err)
        return None

    return destination / release_root_dir


def _update_config(config_path: Path, key: str, value: str) -> None:
    """Modify the configuration at a given path for key to hold value"""
    with open(config_path, mode='r', encoding='utf-8') as config_file:
        config = config_file.readlines()

    done = False
    directive = f"{key}=\"{value}\"\n"

    for index, line in enumerate(config):
        if not done and line.startswith(key):
            config[index] = directive
            done = True

    if not done:
        config.append(directive)

    with open(config_path, mode='w', encoding='utf-8') as config_file:
        config_file.writelines(config)


def overwrite_config(config_path: Path, key: str, value: str) -> None:
    """Make sure the only defined key in the configuration is the one passed as
    an argument"""
    for vendor in WI4MPI_METADATA:
        _update_config(config_path, vendor.default_path_key, '')
    _update_config(config_path, key, value)


def _double_tap(cmd):
    """
    Run a given command (cmake/make) discarding the output. If the
    returncode indicates an error, run it again with the out/err streams
    enabled; this ensures a concise error output as recommended on the GNU
    make's website.
    """

    LOGGER.debug("Double tapping: %s", " ".join(cmd))
    success = run_subprocess(cmd, discard_output=True)
    if success:
        LOGGER.debug("Command run failed: %s, running with visible output.",
                     cmd)
        run_subprocess(cmd, discard_output=False)

    return not success


def _check_wi4mpi_install(install_dir: Path) -> bool:
    """
    Sanity check for the Wi4MPI installation
    """
    # Check binary existence
    binary = install_dir / 'bin' / 'wi4mpi'
    if not binary.exists():
        LOGGER.debug("Wi4MPI binary not found at %s", binary)
        return False

    # Check architecture of shares objects
    # Expected machine from os.uname().machine
    machine = os.uname().machine
    expected = {
        'x86_64': 'EM_X86_64',
        'amd64': 'EM_X86_64',
        'aarch64': 'EM_AARCH64',
        'ppc64le': 'EM_PPC64',
        'ppc64': 'EM_PPC64',
    }.get(machine)

    if expected:
        # Check libexec shared objects
        lib_dir = install_dir / 'libexec' / 'wi4mpi'
        libraries = list(lib_dir.glob('libwi4mpi_*.so'))

        if libraries:
            for lib in libraries:
                if not lib.is_file():
                    continue
                try:
                    with open(lib, 'rb') as f:
                        elf = ELFFile(f)
                        # Note: With pyelftools 0.27, header['e_machine'] returns a string like 'EM_X86_64'
                        # Both elf.header.e_machine and elf.header['e_machine'] work equivalently
                        machine_type = elf.header['e_machine']
                        if machine_type != expected:
                            LOGGER.error(
                                "Wi4MPI library %s architecture mismatch: expected %s, got %s",
                                lib.name, expected, machine_type)
                            return False
                except Exception as err:
                    LOGGER.debug("Failed to check architecture of %s: %s", lib,
                                 err)

    # Check if executable runs
    if run_subprocess([str(binary), '-h'], discard_output=True):
        LOGGER.error("Wi4MPI binary at %s failed to run", binary)
        return False

    return True


def install_wi4mpi(install_dir: Path) -> Optional[Path]:
    """Clones and installs wi4mpi from github releases"""

    if os.uname().machine not in {'x86_64', 'amd64', 'aarch64', 'ppc64le'}:
        LOGGER.warning(
            "Wi4MPI not available for the following architecture: %s",
            os.uname().machine)
        return None

    if install_dir.exists():
        if _check_wi4mpi_install(install_dir):
            LOGGER.debug(
                "Skipping installation for already installed Wi4MPI in %s",
                install_dir)
            return install_dir

        if list(install_dir.glob('*')):
            LOGGER.error(
                "Target directory %s is not empty and contains an invalid or incomplete Wi4MPI installation.",
                install_dir)
            return None

    # Assert CMake is available
    cmake_executable = which("cmake")
    if not cmake_executable:
        LOGGER.error("Wi4MPI installation failed: cmake is missing.")
        return None

    compiler_id = _select_compiler()
    compiler = _WI4MPI_COMPILER_STRINGS.get(compiler_id)
    if compiler is None:
        LOGGER.error("No available compiler to build Wi4MPI: aborting.")
        return None

    source_dir = _download_wi4mpi(WI4MPI_DIR)
    if source_dir is None:
        LOGGER.error("Failed to download Wi4MPI release; aborting")
        return None

    build_dir = WI4MPI_DIR / 'build'
    c_compiler, cxx_compiler, fortran_compiler = VENDOR_BINARIES.get(
        compiler_id)

    configure_cmd = [
        cmake_executable,
        "-B",
        str(build_dir),
        "-S",
        str(source_dir),
        f"-DCMAKE_INSTALL_PREFIX={install_dir}",
        f"-DCMAKE_C_COMPILER={c_compiler}",
        f"-DCMAKE_CXX_COMPILER={cxx_compiler}",
        f"-DCMAKE_FC_COMPILER={fortran_compiler}",
        f"-DWI4MPI_COMPILER={compiler}",
        source_dir.as_posix(),
    ]

    # Apply custom CFLAGS/CXXFLAGS if provided via environment variables
    wi4mpi_cflags = os.environ.get('E4S_CL_WI4MPI_CFLAGS')
    wi4mpi_cxxflags = os.environ.get('E4S_CL_WI4MPI_CXXFLAGS')
    if wi4mpi_cflags:
        configure_cmd.append(f"-DCMAKE_C_FLAGS={wi4mpi_cflags}")
        LOGGER.debug("Using custom Wi4MPI CFLAGS: %s", wi4mpi_cflags)
    if wi4mpi_cxxflags:
        configure_cmd.append(f"-DCMAKE_CXX_FLAGS={wi4mpi_cxxflags}")
        LOGGER.debug("Using custom Wi4MPI CXXFLAGS: %s", wi4mpi_cxxflags)

    build_cmd = [
        cmake_executable,
        "--build",
        str(build_dir),
        '--parallel',
        str(CPU_COUNT),
    ]

    install_cmd = [
        cmake_executable,
        '--build',
        str(build_dir),
        '--target',
        'install',
    ]

    try:
        if build_dir.exists():
            rmtree(build_dir)

        build_dir.mkdir(exist_ok=True)
    except PermissionError as err:
        LOGGER.debug("Failed to create build directory %s: %s",
                     build_dir.as_posix(), str(err))
        return None

    LOGGER.info("Installing Wi4MPI in %s", install_dir)

    if _double_tap(configure_cmd) \
            and _double_tap(build_cmd) \
            and _double_tap(install_cmd):
        LOGGER.info("Wi4MPI has been built and installed")
        return install_dir

    LOGGER.error("Wi4MPI installation failed: MPI translation may fail.")
    rmtree(install_dir, ignore_errors=True)
    return None



"""
Use python-ptrace to catch all open and openat system calls, effectively listing
all files accessed by a program
"""


import pathlib
from ptrace.debugger import (PtraceDebugger, ProcessExit, ProcessSignal,
                             NewProcessEvent, ProcessExecution, child)
from ptrace.func_call import FunctionCallOptions
from ptrace.tools import locateProgram
from e4s_cl import logger

LOGGER = logger.get_logger(__name__)


def opened_files(command):
    """
    Use python-ptrace to list open system calls from the command.
    """
    files = []
    debugger = PtraceDebugger()
    command[0] = locateProgram(command[0])

    try:
        pid = child.createChild(command, no_stdout=False, close_fds=False)
    except child.ChildError as err:
        LOGGER.error("Failed to list opened files of %s: %s", command[0],
                     str(err))
        return -1, []

    # Debugger.addProcess also uses logging, setting the level to warning
    # mutes info messages
    bkp_level = logger.LOG_LEVEL
    logger.set_log_level('WARNING')
    process = debugger.addProcess(pid, is_attached=True)
    logger.set_log_level(bkp_level)

    return_code = 0

    def list_syscalls():
        # Access the return code above - Python 3 only
        nonlocal return_code
        process.syscall()

        while debugger:
            # Wait until next syscall enter
            try:
                event = debugger.waitSyscall()
            except ProcessExit as event:
                return_code = event.exitcode
                # When a process exits in a multi-process MPI run, we need to check
                # if there are other active processes remaining. The exiting process
                # may still be in debugger.dict at this point, so check if there are
                # processes other than the one that just exited.
                # Note: while debugger checks len(debugger.list), which tracks active processes.
                # If the loop continues (debugger is still truthy), we can continue tracing.
                continue
            except ProcessSignal as event:
                event.process.syscall(event.signum)
                continue
            except NewProcessEvent:
                continue
            except ProcessExecution as event:
                print(event)
                continue

            # Process syscall enter or exit
            caught_syscall = event.process.syscall_state.event(FunctionCallOptions())
            if caught_syscall and (caught_syscall.result is not None):
                yield caught_syscall

            # Break at next syscall
            event.process.syscall()

    for syscall in list_syscalls():
        if syscall.result < 0:
            continue
        if syscall.name == "open":
            files.append(syscall.arguments[0].getText())
        if syscall.name == "openat":
            files.append(syscall.arguments[1].getText())

    paths = {name.strip("'") for name in files}
    return return_code, [pathlib.Path(p) for p in paths]



"""
Module introducing shifter support
"""

import os
import subprocess
import tempfile
from pathlib import Path
from typing import List
from e4s_cl import logger, CONTAINER_DIR
from e4s_cl.util import run_subprocess, path_contains
from e4s_cl.cf.containers import Container, FileOptions, BackendNotAvailableError

LOGGER = logger.get_logger(__name__)

NAME = 'shifter'
MIMES = []

OPTION_STRINGS = {FileOptions.READ_ONLY: 'ro', FileOptions.READ_WRITE: 'rw'}

_DEFAULT_CONFIG_PATH = Path('/etc/shifter/udiRoot.conf')


def _deprettify(lines):
    """
    Reconstruct full directives out of directives separated by backslashes
    over multiple lines
    """
    buffer, full_lines = '', []

    for line in lines:
        buffer = buffer + line

        if buffer.endswith('\\'):
            buffer = buffer.rstrip('\\')
            continue

        full_lines.append(buffer)
        buffer = ''

    return full_lines


def _directives_to_dict(directives):
    """
    Transform a list of strings of shape 'KEY=value[=foo]' into a dict
    """
    entries = []

    # Split all the directives at the first '='
    for directive in directives:
        parts = directive.split('=', 1)

        if len(parts) != 2:
            LOGGER.debug("Shifter: udiRoot.conf: Unrecognized directive: '%s'",
                         directive)
            continue

        entries.append(tuple(map(lambda x: x.strip(), parts)))

    return dict(entries)


def _parse_config(config_file: Path):
    """
    Parse a file at a given path and return a dict with its defined variables
    """
    # Read the given file
    try:
        with open(config_file, 'r', encoding='utf-8') as config:
            config_directives = _deprettify(
                [l.strip() for l in config.readlines()])
    except IOError as err:
        LOGGER.warning("Error opening configuration file: %s", str(err))
        return {}

    # Remove all comments
    config_directives = set(
        filter(lambda x: not x.startswith('#'), config_directives))

    # Organize the results in a dict
    return _directives_to_dict(config_directives)


class ShifterContainer(Container):
    """
    Class to use for a shifter execution
    """

    executable_name = 'shifter'

    @classmethod
    @property
    def linker_path(cls):
        """
        Fetch the LD_LIBRARY_PATH from the configuration file
        """
        config = _parse_config(_DEFAULT_CONFIG_PATH)

        path = []

        for module in config.get('defaultModules', '').split(','):
            prepend = config.get(f"module_{module}_siteEnvPrepend")
            if prepend:
                for var in prepend.split():
                    if var.startswith('LD_LIBRARY_PATH'):
                        path.append(var.split('=')[1])

        return os.pathsep.join(path).split(os.pathsep)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.temp_dir = None

    def _setup_import(self, where: Path) -> str:
        """
        Create a temporary directory to bind /.e4s-cl files in
        """
        volumes = [(where.as_posix(), CONTAINER_DIR)]

        for file in self.bound:
            if path_contains(Path('/var'), file.destination):
                LOGGER.debug("Omitting bind of %s to %s: forbidden bind path",
                             str(file.origin), str(file.destination))
                continue

            if file.destination.as_posix().startswith(CONTAINER_DIR):
                rebased = file.destination.as_posix()[len(CONTAINER_DIR) + 1:]
                temporary = Path(where, rebased)

                LOGGER.debug("Shifter: Creating %s for %s in %s",
                             temporary.as_posix(), file.origin.as_posix(),
                             file.destination.as_posix())
                os.makedirs(temporary.parent, exist_ok=True)
                with subprocess.Popen(
                    ['cp', '-r',
                     file.origin.as_posix(),
                     temporary.as_posix()]) as proc:
                    proc.wait()

            elif file.origin.is_dir():
                if file.destination.as_posix().startswith('/etc'):
                    LOGGER.error(
                        "Shifter: Backend does not support binding to '/etc'")
                    continue

                volumes.append(
                    (file.origin.as_posix(), file.destination.as_posix()))

            else:
                LOGGER.warning(
                    "Shifter: Failed to bind '%s': Backend does not support file"
                    "binding. Performance may be impacted.", file.origin)

        return [f"--volume={source}:{dest}" for (source, dest) in volumes]

    def _prepare(self, command: List[str], overload: bool = True) -> List[str]:
        env_list = []
        if self.ld_preload:
            env_list.append(f'--env=LD_PRELOAD={":".join(self.ld_preload)}')
        if self.ld_lib_path:
            env_list.append(
                f'--env=LD_LIBRARY_PATH={":".join(self.ld_lib_path)}')

        for env_var in self.env.items():
            env_list.append(f'--env={env_var}={env_var}')

        # The following is a variable linked to a directory created on the disk
        # Erasing this variable will erase the directory, thus the bind to self
        # pylint: disable=R1732
        self.temp_dir = tempfile.TemporaryDirectory()
        volumes = self._setup_import(Path(self.temp_dir.name))
        return [
            f"--image={self.image}",
            *env_list,
            *volumes,
            *self._additional_options(),
            *command,
        ]

    def run(self, command: List[str], overload: bool = True) -> int:
        executable = self._executable()
        if executable is None:
            raise BackendNotAvailableError(self.__class__.__name__)

        container_cmd = [executable, *self._prepare(command, overload)]
        LOGGER.debug(container_cmd)
        return run_subprocess(container_cmd, env=self.env)


CLASS = ShifterContainer



import os
import tempfile
import tarfile
from pathlib import Path
import tests
from e4s_cl.util import which, path_accessible, safe_tar


class UtilTest(tests.TestCase):

    def test_which(self):
        executable = Path(which('ls'))
        self.assertTrue(executable.exists())
        self.assertTrue(executable.is_absolute())
        return executable.as_posix()

    @tests.skipIf("CICD" in os.environ, "gitlab testing environment")
    def test_access(self):
        self.assertTrue(path_accessible('/tmp', 'r'))
        self.assertTrue(path_accessible('/tmp', 'w'))
        self.assertTrue(path_accessible('/tmp', 'x'))
        self.assertTrue(path_accessible('/tmp', 'rwx'))
        self.assertFalse(path_accessible('/tmp/nonexistentfilehellothere',
                                         'w'))
        self.assertFalse(path_accessible('/root', 'w'))

    @tests.skipIf(True,
                  "Test hangs for no reason, disabling until fix is found")
    def test_safe_tar(self):
        with tempfile.NamedTemporaryFile(mode='w+b', suffix='.tgz') as archive:
            # Shape /home/foo/bar
            with tarfile.open(archive.name, 'w') as absolute:
                absolute_file = tarfile.TarInfo(str(Path.home() / '.bashrc'))
                absolute.addfile(absolute_file)
                self.assertFalse(safe_tar(absolute))

            # Shape ../foo/bar
            with tarfile.open(archive.name, 'w') as relative:
                relative_file = Path('..') / Path.cwd().name

                relative.add(relative_file, arcname=str(relative_file))
                self.assertFalse(safe_tar(relative))

            # Shape foo/../foo/bar
            with tarfile.open(archive.name, 'w') as safe_relative:
                current_path = Path.cwd()
                relative_file = Path(current_path.parts[0], '..', current_path)
                safe_relative.add(relative_file, arcname=str(relative_file))
                self.assertTrue(safe_tar(safe_relative))



"""
Definition of the help command. Subject to change as it was mostly imported from taucmdr.
"""

import os
import mimetypes
from e4s_cl import EXIT_SUCCESS, E4S_CL_SCRIPT
from e4s_cl import logger, util, cli
from e4s_cl.cli import arguments, UnknownCommandError
from e4s_cl.cli.command import AbstractCommand

LOGGER = logger.get_logger(__name__)

_SCRIPT_CMD = os.path.basename(E4S_CL_SCRIPT)

_GENERIC_HELP = f"See '{_SCRIPT_CMD} --help' or raise an issue on Github for assistance"

_KNOWN_FILES = {}

def _fuzzy_index(dct, full_key):
    """Return d[key] where ((key in k) == true) or return d[None]."""
    for key in dct:
        if key and (key in full_key):
            return dct[key]
    return dct[None]


def _guess_filetype(filename):
    """Return a (filetype, encoding) tuple for a file."""
    mimetypes.init()
    filetype = mimetypes.guess_type(filename)
    if not filetype[0]:
        textchars = bytearray([7, 8, 9, 10, 12, 13, 27]) + bytearray(
            range(0x20, 0x100))
        with open(filename, encoding='utf-8') as fd:
            if fd.read(1024).translate(None, textchars):
                filetype = ('application/unknown', None)
            else:
                filetype = ('text/plain', None)
    return filetype


def _gen_parts(cmd_obj):
    parts = []

    if cmd_obj.help_page:
        parts.extend(
            ["", util.hline("Help: " + cmd_obj.command), cmd_obj.help_page])
    if cmd_obj.usage:
        parts.extend(
            ["", util.hline("Usage: " + cmd_obj.command), cmd_obj.usage])

    return parts


class HelpCommand(AbstractCommand):
    """``help`` subcommand."""
    @staticmethod
    def exit_with_help(name):
        """Show a subcommands help page and exit."""
        cmd_obj = cli.find_command(name)
        util.page_output('\n'.join(_gen_parts(cmd_obj)))
        return EXIT_SUCCESS

    @staticmethod
    def exit_with_fullhelp():
        """Show a recursive help page for all commands and exit."""
        help_output = ''
        for cmd_name in cli.get_all_commands():
            name = cli.command_from_module_name(cmd_name)
            cmd_obj = cli.find_command(name.split()[1:])
            help_output += '\n'.join(_gen_parts(cmd_obj))
        util.page_output(help_output)
        return EXIT_SUCCESS

    def _construct_parser(self):
        usage_head = f"{self.command} <command>|<file>|all [arguments]"
        parser = arguments.get_parser(prog=self.command,
                                      usage=usage_head,
                                      description=self.summary)
        parser.add_argument('command',
                            help="An E4S command, system command, or file.",
                            metavar='(<command>|<file>|all)',
                            nargs='+')
        return parser

    def main(self, argv):
        args = self._parse_args(argv)
        if not args.command:
            return self.exit_with_help([])
        if args.command[0] == 'all':
            return self.exit_with_fullhelp()

        # Try to look up a e4s-cl command's built-in help page
        cmd = args.command
        try:
            return self.exit_with_help(cmd)
        except ImportError:
            pass

        # Is this a file?
        if os.path.exists(cmd):
            # Do we recognize the file name?
            try:
                desc, hint = _fuzzy_index(_KNOWN_FILES, cmd.lower())
            except KeyError:
                pass
            else:
                article = 'an' if desc[0] in 'aeiou' else 'a'
                hint = f"'{cmd}' is {article} {desc}.\n{hint}."
                raise UnknownCommandError(cmd, hint)

            # Get the filetype and try to be helpful.
            filetype, encoding = _guess_filetype(cmd)
            self.logger.debug("'%s' has filetype (%s, %s)", cmd, filetype,
                              encoding)
            raise UnknownCommandError(cmd)

        LOGGER.error("Cannot identify '%s' as a command or filename.", cmd)
        return self.exit_with_help('__main__')


COMMAND = HelpCommand(
    __name__,
    summary_fmt="Show help for a command or suggest actions for a file.")



"""
Controller definition of the MVC architecture
"""

from e4s_cl import logger
from e4s_cl.error import UniqueAttributeError, ModelError

LOGGER = logger.get_logger(__name__)

# Suppress debugging messages in optimized code
if __debug__:
    _heavy_debug = LOGGER.debug  # pylint: disable=invalid-name
else:

    def _heavy_debug(*args, **kwargs):
        # pylint: disable=unused-argument
        pass


class Controller():
    """The "C" in `MVC`_.

    Attributes:
        model (AbstractModel): Data model.
        storage (AbstractDatabase): Record storage. 
    
    .. _MVC: https://en.wikipedia.org/wiki/Model-view-controller
    """

    messages = {}

    def __init__(self, model_cls, storage):
        self.model = model_cls
        self.storage = storage

    @classmethod
    def push_to_topic(cls, topic, message):
        cls.messages.setdefault(topic, []).append(message)

    @classmethod
    def pop_topic(cls, topic):
        return cls.messages.pop(topic, [])

    def one(self, key):
        """Get a record.
        
        Args:
            key: See :any:`AbstractStorage.get`.
            
        Returns:
            Model: The model for the matching record or None if no such record exists.
        """
        record = self.storage.get(key, table_name=self.model.name)
        return self.model(record) if record else None

    def all(self):
        """Get all records.
        
        Returns:
            list: Models for all records or an empty lists if no records exist.
        """
        return [
            self.model(record)
            for record in self.storage.search(table_name=self.model.name)
        ]

    def count(self):
        """Return the number of records.
        
        Returns:
            int: Effectively ``len(self.all())``
        """
        return self.storage.count(table_name=self.model.name)

    def search(self, keys=None):
        """Return records that have all given keys.
        
        Args:
            keys: See :any:`AbstractStorage.search`.
            
        Returns:
            list: Models for records with the given keys or an empty lists if no records have all keys.
        """
        return [
            self.model(record)
            for record in self.storage.search(keys=keys,
                                              table_name=self.model.name)
        ]

    def match(self, field, regex=None, test=None):
        """Return records that have a field matching a regular expression or test function.
        
        Args:
            field: See :any:`AbstractStorage.match`.
            regex: See :any:`AbstractStorage.match`.
            test: See :any:`AbstractStorage.match`.
            
        Returns:
            list: Models for records that have a matching field.
        """
        return [
            self.model(record) for record in self.storage.match(
                field, table_name=self.model.name, regex=regex, test=test)
        ]

    def exists(self, keys):
        """Check if a record exists.
        
        Args:
            keys: See :any:`AbstractStorage.exists`.
            
        Returns:
            bool: True if a record matching `keys` exists, False otherwise.
        """
        return self.storage.contains(keys, table_name=self.model.name)

    def _check_unique(self, data, match_any=True):
        unique = {
            attr: data[attr]
            for attr, props in self.model.attributes.items()
            if 'unique' in props
        }
        if unique and self.storage.contains(
                unique, match_any=match_any, table_name=self.model.name):
            raise UniqueAttributeError(self.model, unique)

    def create(self, data):
        """Atomically store a new record
        
        Invokes the `on_create` callback **after** the data is recorded.  If this callback raises
        an exception then the operation is reverted.
        
        Args:
            data (dict): Data to record.
            
        Returns:
            Model: The newly created data. 
        """
        data = self.model.validate(data)
        self._check_unique(data)
        with self.storage as database:
            record = database.insert(data, table_name=self.model.name)
            model = self.model(record)
            model.check_compatibility(model)
            model.on_create()
            return model

    def update(self, data, keys):
        """Change recorded data
        
        The behavior depends on the type of `keys`:
            * Record.ElementIdentifier: update the record with that element identifier.
            * dict: update all records with attributes matching `keys`.
            * list or tuple: apply update to all records matching the elements of `keys`.
            * ``bool(keys) == False``: raise ValueError.
            
        Invokes the `on_update` callback **after** the data is modified.  If this callback raises
        an exception then the operation is reverted.

        Args:
            data (dict): New data for existing records.
            keys: Fields or element identifiers to match.
        """
        old_records = self.search(keys)

        if not old_records:
            raise ModelError(self.model,
                             f"No matching models for query {keys}")

        for attr, value in data.items():
            if attr not in self.model.attributes:
                raise ModelError(self.model, f"no attribute named '{attr}'")

            if attr == self.model.key_attribute:
                if len(old_records) > 1:
                    raise ModelError(
                        self.model,
                        f"Updating {len(old_records)} {self.model.name.lower()}s"
                        f" with {attr}={value} would break unicity rules")

                existing_record = self.one({attr: value})
                if existing_record and existing_record not in old_records:
                    raise UniqueAttributeError(self.model, {attr: value})

        with self.storage as database:
            # Get the list of affected records **before** updating the data so foreign keys are correct
            database.update(data, keys, table_name=self.model.name)
            changes = {}
            for model in old_records:
                changes[model.eid] = {
                    attr: (model.get(attr), new_value)
                    for attr, new_value in data.items()
                    if not (attr in model and model.get(attr) == new_value)
                }
            updated_records = self.search(keys)
            for model in updated_records:
                model.check_compatibility(model)
                model.on_update(changes[model.eid])

    def unset(self, fields, keys):
        """Unset recorded data fields
        
        The behavior depends on the type of `keys`:
            * Record.ElementIdentifier: update the record with that element identifier.
            * dict: update all records with attributes matching `keys`.
            * list or tuple: apply update to all records matching the elements of `keys`.
            * ``bool(keys) == False``: raise ValueError.

        Invokes the `on_update` callback **after** the data is modified.  If this callback raises
        an exception then the operation is reverted.

        Args:
            fields (list): Names of fields to unset.
            keys: Fields or element identifiers to match.
        """
        for attr in fields:
            if attr not in self.model.attributes:
                raise ModelError(self.model, f"no attribute named '{attr}'")
        with self.storage as database:
            # Get the list of affected records **before** updating the data so foreign keys are correct
            old_records = self.search(keys)
            database.unset(fields, keys, table_name=self.model.name)
            changes = {}
            for model in old_records:
                changes[model.eid] = {
                    attr: (model.get(attr), None)
                    for attr in fields if attr in model
                }
            updated_records = self.search(keys)
            for model in updated_records:
                model.check_compatibility(model)
                model.on_update(changes[model.eid])

    def delete(self, keys):
        """Delete recorded data
        
        The behavior depends on the type of `keys`:
            * Record.ElementIdentifier: delete the record with that element identifier.
            * dict: delete all records with attributes matching `keys`.
            * list or tuple: delete all records matching the elements of `keys`.
            * ``bool(keys) == False``: raise ValueError.

        Invokes the `on_delete` callback **after** the data is deleted.  If this callback raises
        an exception then the operation is reverted.

        Args:
            keys (dict): Attributes to match.
            keys: Fields or element identifiers to match.
        """
        with self.storage as database:
            removed_data = []
            changing = self.search(keys)
            for model in changing:
                removed_data.append(dict(model))
            database.remove(keys, table_name=self.model.name)
            for model in changing:
                model.on_delete()

    @staticmethod
    def import_records(data):
        """Import data records.
        
        TODO: Docs
        """

    @classmethod
    def export_records(cls, keys=None, eids=None):
        """Export data records.
        
        Constructs a dictionary containing records matching `keys` or `eids`

        Args:
            keys (dict): Attributes to match.
            eids (list): Record identifiers to match.

        Returns:
            dict: Dictionary of tables containing records.
            
        Example:
        ::
            
            {
             'Brewery': {100: {'address': '4615 Hollins Ferry Rd, Halethorpe, MD 21227',
                               'brews': [10, 12, 14]}},
             'Beer': {10: {'origin': 100, 'color': 'gold', 'ibu': 45},
                      12: {'origin': 100, 'color': 'dark', 'ibu': 15},
                      14: {'origin': 100, 'color': 'pale', 'ibu': 30}}
            }
        
            Beer.export_records(eids=[10])
            
            {
             'Brewery': {100: {'address': '4615 Hollins Ferry Rd, Halethorpe, MD 21227',
                               'brews': [10, 12, 14]}},
             'Beer': {10: {'origin': 100, 'color': 'gold', 'ibu': 45}}
            }

        """

        def export_record(record, root):
            if isinstance(record, cls) and record is not root:
                return
            data = all_data.setdefault(record.model_name, {})
            if record.eid not in data:
                data[record.eid] = record.data

        all_data = {}
        for record in cls.search(keys, eids):
            export_record(record, record)
        return all_data



"""
Functions used in the e4s-cl-mpi-tester script
"""

import ctypes
import logging
import sys
from typing import Optional
from pathlib import Path
from argparse import ArgumentParser
from sotools.linker import resolve

DESCRIPTION = """This script will dynamically load the MPI library passed as an \
        argument and run a simple program with it. This serves no purpose by \
        itself and is meant to be traced to list the files an arbitrary MPI \
        library requires to run. If no library is given, the first MPI library \
        found in the environment will be used."""
EPILOG = """Please report any issues to http://github.com/E4S-Project/e4s-cl."""

PARSER = ArgumentParser(
    prog=sys.argv[0],
    description=DESCRIPTION,
    epilog=EPILOG,
)

PARSER.add_argument(
    "library",
    nargs='?',
    default=None,
    help="The MPI shared object to analyze.",
)

PARSER.add_argument(
    "-n",
    "--no-mpi",
    action="store_true",
    help="Disable MPI binding; run an empty program",
)

PARSER.add_argument(
    "-v",
    "--verbose",
    action="store_true",
    help="Trace resolving attempts while searching for the library",
)


class MPIHandles:  #pylint: disable=too-few-public-methods
    """
    Empty class to which MPI function/variable bindings are added as attributes.
    """


def bind_mpich(lib: ctypes.CDLL) -> MPIHandles:
    """
    Add bindings from an MPICH library to a MPIHandles object and return it
    """
    handles = MPIHandles()
    setattr(handles, "Init", getattr(lib, "MPI_Init"))
    setattr(handles, "Comm_size", getattr(lib, "MPI_Comm_size"))
    setattr(handles, "Comm_rank", getattr(lib, "MPI_Comm_rank"))
    setattr(handles, "Get_processor_name",
            getattr(lib, "MPI_Get_processor_name"))
    setattr(handles, "Finalize", getattr(lib, "MPI_Finalize"))
    setattr(handles, "Barrier", getattr(lib, "MPI_Barrier"))
    setattr(handles, "Allreduce", getattr(lib, "MPI_Allreduce"))
    setattr(handles, "COMM_WORLD", ctypes.c_int(0x44000000))
    setattr(handles, "FLOAT", ctypes.c_int(0x4c00040a))
    setattr(handles, "SUM", ctypes.c_int(0x58000003))
    setattr(handles, "MAX_PROCESSOR_NAME", 128)

    return handles


def bind_ompi(lib: ctypes.CDLL) -> MPIHandles:
    """
    Add bindings from an OpenMPI library to a MPIHandles object and return it
    """
    handles = MPIHandles()
    setattr(handles, "Init", getattr(lib, "MPI_Init"))
    setattr(handles, "Comm_size", getattr(lib, "MPI_Comm_size"))
    setattr(handles, "Comm_rank", getattr(lib, "MPI_Comm_rank"))
    setattr(handles, "Get_processor_name",
            getattr(lib, "MPI_Get_processor_name"))
    setattr(handles, "Finalize", getattr(lib, "MPI_Finalize"))
    setattr(handles, "Barrier", getattr(lib, "MPI_Barrier"))
    setattr(handles, "Allreduce", getattr(lib, "MPI_Allreduce"))
    setattr(handles, "COMM_WORLD", getattr(lib, "ompi_mpi_comm_world"))
    setattr(handles, "FLOAT", getattr(lib, "ompi_mpi_float"))
    setattr(handles, "SUM", getattr(lib, "ompi_mpi_op_sum"))
    setattr(handles, "MAX_PROCESSOR_NAME", 256)

    return handles


SONAMES = [
    ("libmpi.so.12", bind_mpich),
    ("libmpi.so.0.0.0", bind_mpich),
    ("libmpi.so.40", bind_ompi),
    ("libmpi.so.41", bind_ompi),
    ("libmpi_cray.so.12", bind_mpich),
]


def select_bind_library() -> Optional[MPIHandles]:
    for (name, bindgen) in SONAMES:
        path = resolve(name)

        if path:
            libhandle = ctypes.CDLL(path)
            logging.info("Using library '%s'", path)
            return bindgen(libhandle)

    return None


def bind_library(path: Path) -> Optional[MPIHandles]:
    for (name, bindgen) in SONAMES:
        if path.resolve().name.startswith(name):
            libhandle = ctypes.CDLL(path)
            return bindgen(libhandle)

    return None


def main():
    """
    Main routine of the e4s-cl-mpi-tester script
    """
    #pylint: disable=no-member
    args = PARSER.parse_args()

    if args.verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(message)s",
        )

    if args.no_mpi:
        libc = resolve("libc.so.6")
        ctypes.CDLL(libc)
        return 0

    try:
        #pylint: disable=invalid-name
        if args.library is not None:
            MPI = bind_library(Path(args.library))
        else:
            MPI = select_bind_library()
    except AttributeError as err:
        logging.error("Binding creation failed: %s", err)
        sys.exit(1)

    if MPI is None:
        if args.library is not None:
            logging.error("Failed to bind to the MPI library")
        else:
            logging.error("Failed to find an MPI library")
        sys.exit(1)

    world_size = ctypes.c_int()
    world_rank = ctypes.c_int()
    name_len = ctypes.c_int()
    local_sum = ctypes.c_int()
    global_sum = ctypes.c_int()
    processor_name = ctypes.create_string_buffer(MPI.MAX_PROCESSOR_NAME)

    MPI.Init(ctypes.c_int(0), ctypes.c_int(0))
    MPI.Comm_size(MPI.COMM_WORLD, ctypes.byref(world_size))
    MPI.Comm_rank(MPI.COMM_WORLD, ctypes.byref(world_rank))
    MPI.Get_processor_name(processor_name, ctypes.byref(name_len))

    print(
        processor_name.value.decode("utf-8"),
        "-",
        world_rank.value,
        "/",
        world_size.value,
    )

    # Trigger a communication with dummy values to ensure the communication
    # code is loaded, as some libraries do it on the fly
    MPI.Allreduce(
        ctypes.byref(local_sum),
        ctypes.byref(global_sum),
        ctypes.c_int(1),
        MPI.FLOAT,
        MPI.SUM,
        MPI.COMM_WORLD,
    )

    MPI.Barrier(MPI.COMM_WORLD)
    MPI.Finalize()


if __name__ == "__main__":
    main()
